Package: qkiosk
Type: Package
Title: 'QUANTkiosk' API Client
Version: 0.2.31
Imports: curl, jsonlite, stringdist, zoo
Depends: xts
Suggests: fansi
LazyData: yes
Date: 2026-04-22
Authors@R: person(given = c("Jeffrey", "A."), family = "Ryan", role = c("aut", "cre"), email = "jeff@quantkiosk.com", comment=c(ORCID="0009-0003-3627-7896"))
Maintainer: Jeffrey A. Ryan <jeff@quantkiosk.com>
Description: Provide access to 'QUANTkiosk' platform of regulatory and corporate data relevant to financial markets including point-in-time fundamentals, ownership and symbology.
License: GPL-3
URL: https://quantkiosk.com/, https://github.com/quantkiosk/qkiosk-r
BugReports: https://github.com/quantkiosk/qkiosk-r/issues
NeedsCompilation: no
Packaged: 2026-04-22 15:45:49 UTC; jeffreyryan
Author: Jeffrey A. Ryan [aut, cre] (ORCID:
    <https://orcid.org/0009-0003-3627-7896>)
Repository: CRAN
Date/Publication: 2026-04-24 18:20:02 UTC
Built: R 4.5.2; ; 2026-04-24 18:30:32 UTC; unix

Package: qkiosk
Type: Package
Title: 'QUANTkiosk' API Client
Version: 0.2.31
Imports: curl, jsonlite, stringdist, zoo
Depends: xts
Suggests: fansi
LazyData: yes
Date: 2026-04-22
Authors@R: person(given = c("Jeffrey", "A."), family = "Ryan", role = c("aut", "cre"), email = "jeff@quantkiosk.com", comment=c(ORCID="0009-0003-3627-7896"))
Maintainer: Jeffrey A. Ryan <jeff@quantkiosk.com>
Description: Provide access to 'QUANTkiosk' platform of regulatory and corporate data relevant to financial markets including point-in-time fundamentals, ownership and symbology.
License: GPL-3
URL: https://quantkiosk.com/, https://github.com/quantkiosk/qkiosk-r
BugReports: https://github.com/quantkiosk/qkiosk-r/issues
NeedsCompilation: no
Packaged: 2026-04-22 15:45:49 UTC; jeffreyryan
Author: Jeffrey A. Ryan [aut, cre] (ORCID:
    <https://orcid.org/0009-0003-3627-7896>)
Repository: CRAN
Date/Publication: 2026-04-24 18:20:02 UTC
Built: R 4.5.2; ; 2026-04-24 20:38:28 UTC; unix

import(curl,jsonlite,xts)

importFrom("stats", "setNames", "na.omit")
importFrom("utils", "str", "read.csv", "head", "tail", "unzip", "browseURL", "download.file","menu")
importFrom("xts","as.xts")
importFrom("zoo","index","rollapply")
importFrom("stringdist","stringdist")
importFrom("utils", "capture.output", "read.delim")

export(

  qk_audit,

  qk_select,
  to_df,
  to_ts,

  as.qk_df,

  full,
  highlight,

  # fundamentals
  qk_fncodes,
  qkiditems,
  qk_fn,
  qk_clearcache,

  pit_asfiled,
  pit_asof,

  pitAsFiled,
  pitAsOf,

  # account
  qk_account,
  qk_download,
  qk_set_apikey,
  qk_get_apikey,

  # ownership
  qk_fund,
  qk_insider,
  qk_activist,
  qk_beneficial,
  qk_institutional,
  qk_holders,

  # bulk data
  qk_complete,

  # universe
  qk_univ,

  # qkid
  .qkidVersion,

  qk_search_mgr,
  qk_search_co,
  qkid,
  to_qkid,
  is_qkid,
  as.qkid,
  is.qkid,

  detail,
  classname,

  entity,
  cls,
  instrument,

  qk_ticker, to_ticker,
  qk_cik, to_cik,
  qk_permid, to_permid,
  qk_figi, to_figi,
  qk_name, to_name,
  qk_lei, to_lei,
  qk_isin, to_isin,
  qk_cusip, to_cusip,
  #qk_sector, to_sector,
  #qk_sic, to_sic,

  # utilities
  qk_sort,
  qk_filter,
  qk_summary,

  today,
  yyyymmdd,
  is.yyyymmdd,
  as.yyyymmdd,
  seq.yyyymmdd,
  qtrs,
  days,

  to_yyyy,
  to_qq,
  to_yyyyqq
)
  
S3method(detail,default)
S3method(detail,qkid)

S3method(qk_audit, default)
S3method(print, qkaudit)

S3method(qk_select, qk_fn)
S3method(print, qk_fn)
S3method(to_df, qk_fn)
S3method(to_df, data.frame)
S3method(to_ts, qk_fn)
S3method(`[`, qk_fn)
S3method(subset, qk_fn)
S3method(transform, qk_fn)
S3method(full, qk_fn)
S3method(as.data.frame, qk_fn)

S3method(highlight, data.frame)
S3method(print, qk_df)
S3method(head, qk_df)
S3method(tail, qk_df)
S3method(full, qk_df)
S3method(`[`, qk_df)
S3method(to_ts, qk_df)
S3method(to_df, qk_df)
S3method(subset, qk_df)
S3method(transform, qk_df)

S3method(print, qk_ts)
S3method(as.xts, qk_ts)

S3method(print, qkid_item)
S3method(str, qkid_item)

S3method(`[`, qkid)
S3method(length, qkid)
S3method(c, qkid)
S3method(as.character, qkid)
S3method(print, qkid)
S3method(print, qkid_detail)

S3method(print, qkaccount)

S3method(qk_sort, qk_df)
S3method(qk_filter, qk_df)
S3method(qk_summary, qk_df)
S3method(print, qk_df_fn)
S3method(print, qk_df_institutional)
S3method(print, qk_df_holders)
S3method(print, qk_df_insider)
S3method(print, qk_df_beneficial)

S3method(as.yyyymmdd, yyyymmdd)
S3method(as.yyyymmdd, numeric)
S3method(as.yyyymmdd, Date)
S3method(as.Date, yyyymmdd)
S3method(as.POSIXct, yyyymmdd)
S3method(as.POSIXlt, yyyymmdd)
S3method(Ops, yyyymmdd)
S3method(print, yyyymmdd)
S3method(seq, yyyymmdd)
S3method(diff, yyyymmdd)
S3method(to_qq, yyyymmdd)
S3method(to_yyyy, yyyymmdd)
S3method(to_yyyyqq, yyyymmdd)

<img src="https://quantkiosk.com/assets/img/qk-logo.png" height="40" />

>[!IMPORTANT]
> **FREE API keys are now enabled for all accounts. Get yours [here](https://quantkiosk.com?ref=github-qkiosk-r)!!**
>
> Paid plans will be open in the coming weeks as we finish backfill and add more features!

## R Client

Official interface to [QUANTkiosk](https://quantkiosk.com) data api.

<img width="753" alt="image" src="https://github.com/user-attachments/assets/5d328e70-cb5a-4a8e-b0f2-2bf4d6de8d72" />

### Features
* A true extension of `R` - use the tools you are productive with.
* Integrated symbology to easily map entities and instruments.
* Limited dependencies to ensure easy installation and no conflicts.
* Access to all data endpoints, including Ownership and Fundamentals.
* Parallel API requests for large downloads in a fraction of the time.
* Internal caching to minimize external requests and data usage

### Installation
```r
# make sure you have the "remotes" package installed, if not
# you need to run install.packages("remotes") first.
library(remotes)

install_github("https://github.com/quantkiosk/qkiosk-r")
```

### Set up your API key
All access to live and historical data requires a valid `QK_API_KEY` to be set. To get your
**FREE** key go to the [account page](https://www.quantkiosk.com/account). If you have not signed up for an account, you can
enter your email and your account key will be good for 250 credits a day. More than enough to explore and
make use of the API. If you need more data, just select an appropriate plan.

```r
## Set your API key in the R session after you've load the freshly installed package

require(qkiosk)
qk_set_apikey("<YOUR_API_KEY>")
```

>[!TIP]
>```bash
>## you can also set up your key in your shell to avoid having to set it in R
>## this is more permanent and definitely how you would do it in production
>
>export QK_API_KEY=<YOUR_API_KEY>
>```

>[!TIP]
>You don't actually need a key to get started with the package. We've included datasets that represent what each of the endpoints
>return to help get a feel for the breadth and depth of what QK does.
>```
>data(package="qkiosk")
># Data sets in package ‘qkiosk’:
>#
># crox        Crocs Institutional Holders Details By Issuer
># deshaw      D.E. Shaw Institutional Ownership Details (Including Submanagers)
># nke         Nike (NKE) Insider Ownership Data
># nvda_intent Nvidia (NVDA) Restricted or Control Shares Intention to Sell
># nvda_sales  Nvidia (NVDA) Restricted or Control Shares Sold
># pershing    Pershing Square Beneficial and Activist Details
># pfe         Pfizer (PFE) Revenue Data
># sgcap       SG Capital Institutional Ownership Details (Aggregated)

## Get Started

### Symbology drives everything.

The most important part of any institutional data is getting the mappings correct. The best hedge funds take having a security master for granted. It is also
an incredible pain point that firms spend millions a year to manage, or tens of thousands (or more!) to buy.

Once you have access to a proper symbology - you won't understand how you ever lived without it.

**QUANTkiosk** is solving this once and for all by creating an open *entity and security* master to help
leverage our data with as little effort as possible.  You can read more about our project on the site, but in short, everything
within the API is referenced by a **QKID**. To make that _easy_ we have some helper functions. Both search and conversion
is part of the batteries included mindset of QK.

Obviously you likely have _some_ identifier to begin with. Most often this is a ticker, like **AAPL** or **MRK**. You can use this
to map to a **QKID**, but you can even search right from your `R` session. This is especially useful for things that
don't have ticker - e.g. a hedge fund you want to track.

```
# You know the ticker or cik? Just use direct conversion:
qk_ticker("AAPL")
#                      AAPL 
# 0000320193.0000.001S5N8V8

to_name(qk_cik(78003))
# [1] "PFIZER INC"
to_permid(qk_cik(78003))
# [1] "4295904722"

# use fuzzy search for a company like Alibaba
BABA <- qk_search_co("alibaba")
#
# 1: ALIBABA GROUP HOLDING-SP ADR
# 2: ALITHYA USA INC
# 3: ALIGNMENT HEALTHCARE INC
# 4: ALTI GLOBAL INC
# 5: ALLY FINANCIAL INC
# 6: ALLEGIANT TRAVEL CO
#
#Selection: 1
BABA
# ALIBABA GROUP HOLDING-SP ADR 
#    0001577552.0400.006G2JWB1
```

You can also search for fund managers in a similar way
```
janest <- qk_search_mgr("jane street")

# 1: JANE STREET GROUP LLC
# 2: JANNEY MONTGOMERY SCOTT LLC
# 3: JANA PARTNERS MANAGEMENT LP
# 4: JANNEY CAPITAL MANAGEMENT LLC
# 5: JOURNEY STRATEGIC WEALTH LLC
# 6: JAMES INVESTMENT RESEARCH INC

# Selection: 1

janest
#     JANE STREET GROUP LLC 
# 0001595888.0000.E0000Y7E8 
> 

```

There is *way* more to know about the **QKID**, but this is the README, so we will move along.  Be sure to try conversion tools like `qk_ticker("AAPL")` and `to_ticker(qk_cik(320187))` to see the power
for yourself.

>[!NOTE]
>A **QKID** is actually a just a unique combination of entity, class and an instrument:
>
>  *[ENTITY].[CLS].[INSTRUMENT]*
>
>The entity is most often the **CIK**, the instrument is an OpenFIGI **FIGI**, and the class is something that helps to quickly identify what this
>instrument is.  Pretty simple, but unlike most everything you might have seen in a security master.
>
>Alibaba will serve as an example of what this looks like in practice
>```
> BABA
> # ALIBABA GROUP HOLDING-SP ADR 
> # 0001577552.0400.006G2JWB1
>```
> - *[ENTITY]* the 10 digit CIK assigned by the SEC to any entity that files in the US. Even foreign firms need one at times.
> - *[CLS]* is the classification of the instrument. In this case the 0400 is an ADR tradable in the US. 0000 is used for the common equity - often Class A Ordinary, and there are many others available
> - *[INSTRUMENT]* is the variable portion of the FIGI from OpenFIGI, which is quickly becoming the de-facto identifier - covering a billion+ instruments
>
>For entities that do not have instruments (e.g. a hedge fund or a CEO), we create a unique instrument part to allow for consistency (e.g. **E0000Y7E8** like in Jane Street's case above).

## Get some data.

QK's job is to let you do _your_ job. Mapped, point-in-time, and even auditable - all from within your preferred platform - is how we make that happen. 
To get a feel for this, we'll take a look at two foundational areas - **Ownership** and **Fundamentals**:

- *Universe*
- *Institutional Ownership*
- *Insider Ownership*
- *Activist Ownership*
- *Fundamentals*

Let's get started

### Start with a Universe

In all institutional settings, you are almost always working with a universe. This is nothing more than a collection
of firms or instruments that are part of a strategy. These are often defined by a set of rules - e.g. minimum market cap and
price, some limits on minimum daily volume, etc.

To get you started, we have created a set of universe definitions to showcase the design and data behind **QUANTkiosk**. We'll elaborate on methodologies at another time,
but these represent the most widely held companies amongst institutional investors. From the top 100 to 3000 firms (**QK100**, **QK1000**, **QK3000**, with **QK2000** being those firms less widley held than
the top 1000)

```
qk100 <- qk_univ("QK100")
head(qk100)
# [1] 0000320193.0000.001S5N8V8 0001018724.0000.001S5PQL7 0001652044.0000.009S3NB21 0001045810.0000.001S5TZJ6 0001326801.0000.001SQCQC5 0000019617.0000.001S8CRC3

head(to_ticker(qk100))
#[1] "AAPL" "AMZN" "GOOG" "NVDA" "META" "JPM" 

head(to_name(qk100))
# [1] "APPLE INC"                  "AMAZON.COM INC"             "ALPHABET INC-CL C"          "NVIDIA CORP"                "META PLATFORMS INC-CLASS A" "JPMORGAN CHASE & CO" 
```

Of course, not everything you might want to request is a tradable entity.  In fact, the _owners_ of the above universe are private firms in many cases.

### Institutional Ownership

What they call "smart money". Certain asset managers (those with over $100mm in reportable assets) are obliged to disclose their positions at the end of each quarter. The `qkiosk` package includes two
examples - `deshaw` and `sgcap`, both which provide good documentation on what the structure of the data is. For our purposes, lets just request some data to see how easy it is to get a picture of what held last quarter.

First, well get the manager that became particularly popular in the pandemic. The are a very large multistrat player, who reports for both the asset manager business and the securities business. To get a good feel for
what they are doing for the investing side, it is very useful to be able to see submanager details, which QK does very well.

```
citadel <- qk_search_mgr("citadel")
#  1: CITADEL ADVISORS LLC
#  2: CITADEL INVESTMENT ADVISORY INC.
#  3: CIT BANK NA WEALTH MANAGEMENT
#  4: CITY STATE BANK
#  5: CIC WEALTH LLC
#  6: CITIZENS FINANCIAL GROUP INC/RI
#  7: CITY CENTER ADVISORS LLC
#  8: CITY HOLDING CO
#  9: CITIZENS NATIONAL BANK TRUST DEPARTMENT
# 10: CI PRIVATE WEALTH LLC
# Selection: 1

citadel
#      CITADEL ADVISORS LLC 
# 0001423053.0000.E0000UI19
```
Now that we have our QKID, we can use it in the function called `qk_institutional`. Interface matters a lot to us. If you don't notice how easy it is to use, it is because we've done our job.
```
ca_202402_202501 <- qk_institutional(citadel, yyyyqq=202501, qtrs=4, agg=FALSE)
# fetching 0001423053 for 202501 ...done.
# fetching 0001423053 for 202404 ...done.
# fetching 0001423053 for 202403 ...done.
# fetching 0001423053 for 202402 ...done.
```

![image](https://github.com/user-attachments/assets/ec01fa81-021b-4389-bbc0-84061f6b7590)

It's easy to see what Citadel Securities is holding, how much it has changed, and even see positions they no longer have.
```
ca_202402_202501 |>
  subset(otherManager==1,
         select=c('reportPeriod','issuer','putCall','issuerQkid','issuerSector','value','QOQSshPrnAmt','otherManagerName','newOrDel')) 

```
![image](https://github.com/user-attachments/assets/3206dd38-bfc3-4e34-97ea-38f2068de8f4)

There are of course, a million ways to use this data - and we promise to share videos as well as deep dives as we move forward. We also would love to have anyone using the API share how they 
are discovering insights into what the "Who's Who of Wall Street" are up to each quarter.

We can also explore details of company holders, large shareholders, and even corporate insiders with similar speed and ease.

```
# all holders of NVDA in 2022 Q1
nvda_holders <- qk_holders(qk_ticker("NVDA"), yyyyqq=202201)

# NVDA insiders - (e.g. CEO, Sr.EVP, ...)
nvda_insiders <- qk_insider(qk_ticker("NVDA"), yyyyqq=202201)

# Large NVDA block holders for 2022 (above 5%)
nvda_lg <- qk_beneficial(qk_ticker("NVDA"), yyyyqq=202200)
```

### Fundamentals

Fundamentals data is the lifeblood of a company. While prices are observable, financials are much harder to source
consistently or correctly.  We source directly from filings, in a way only people who have used this data for
investment at scale know how to do.

Let's dig into a quick example to show off what you can do with a few lines of code.
```
# There are hundreds of line items that are generally used - though 10s of 1000s of GAAP items are available.
#
# To see common ones we currently map you can use our reference function qk_fncodes
qk_fncodes()
```
<img width="621" alt="image" src="https://github.com/user-attachments/assets/3993d057-bcaf-4a1d-a696-aaf0931e6398" />

From here, lets find Net Income (code NI) for UBER

```
uber_ni <- qk_fn(qk_ticker("UBER"), "NI")
```
<img width="1051" alt="image" src="https://github.com/user-attachments/assets/06d42814-99dc-4ba7-96de-ea72166bbb93" />

You can also easily see where this data comes from using our state of the art auditing tools
```
qk_fn(qk_ticker("UBER"), "NI") |> to_df() |> tail() |> highlight(5)

qk_fn(qk_ticker("UBER"), "NI") |> to_df() |> tail() |> highlight(5) |> qk_audit()
```
<img width="1022" alt="image" src="https://github.com/user-attachments/assets/3afa85af-168c-4c09-a5f7-8e13635e2d83" />

<img width="1119" alt="image" src="https://github.com/user-attachments/assets/903a6d09-9f5d-416b-a484-449a7504ddc3" />

AAPL.FIGI <- "BBG000B9XRY4"

.QKIDCLASSMAP_V4 <- list(`0000` = "Equity", `000C` = "Equity (Call)", `000P` = "Equity (Put)",
    `0010` = "Warrant", `001C` = "Warrant (Call)", `001P` = "Warrant (Put)",
    `0020` = "Unit", `002C` = "Unit (Call)", `002P` = "Unit (Put)",
    `0030` = "Right", `003C` = "Right (Call)", `003P` = "Right (Put)",
    `0300` = "Convert", `0400` = "ADR", `040C` = "ADR (Call)", `040P` = "ADR (Put)",
    `0500` = "Fund", `050C` = "Fund (Call)", `050P` = "Fund (Put)",
    `0510` = "Closed End Fund", `051C` = "Closed End Fund (Call)", `051P` = "Closed End Fund (Put)",
    `0570` = "REIT", `057C` = "REIT (Call)", `057P` = "REIT (Put)",
    `0600` = "Partnership", `060C` = "Partnership (Call)", `060P` = "Partnership (Put)",
    `0210` = "Note", `0220` = "Debt", `0230` = "Debt (MTN)", `0ZZZ` = "Undefined")
.QKIDREGEXP_V4 <- "([0-9]{10})[.]([A-Z0-9]{4})[.]([BCDFGHJKLMNPQRSTVWXYZ0-9]{9})|([0-9]{10})[.]([A-Z0-9]{4})[.]([UEP][A-Z0-9]{8})"

check_qkid <- function(x) {
  if(!is_qkid(x))
    stop("qkid argument is class '",class(x),"': required to be class qkid. See ?qkid for conversion examples.")
}

.qkidVersion <- function(qkid) {
  if(is.qkid(qkid))
    qkid <- qkid$qkid

  if(all(grepl(.QKIDREGEXP_V4, qkid))) {
    version <- 4
    attr(version, "rx") <- .QKIDREGEXP_V4
  } else {
    version <- NA
  }
  version
} 

qk_search_person <- function(x,n=10) qk_search(x,type='person',n=n)
qk_search_fund <- function(x,n=10) qk_search(x,type='fund',n=n)

qk_search_mgr <- function(x,n=10) qk_search(x,type='manager',n=n)[menu(qk_search(x,type='manager', n=n)$srcid)]
qk_search_co <- function(x,n=10) {
  res <- qk_search(x,type='issuer',n=n)
  res[menu(paste(res$srcid,"\033[38;5;35m",to_ticker(res),"\033[39m"))]
  #[menu(qk_search(x,type='issuer', n=n)$srcid)]
}

qk_search <- function(x, type=c('issuer','manager','fund','person'), n=10) {
  type <- match.arg(type)
  if(type=="manager") {
    .fundnameMapping()
    nm <- grep("^0",names(.QKID_FUNDNAME),invert=TRUE,value=TRUE)
    qk_fundname(nm[order(stringdist(toupper(x), toupper(nm), method='jw', p=.1))[1:n]])
  } else
  if(type=="issuer") {
    .nameMapping()
    nm <- grep("^0",names(.QKID_NAME),invert=TRUE,value=TRUE)
    qk_name(nm[order(stringdist(toupper(x), toupper(nm), method='jw', p=.1))[1:n]])
  } else {
    stop(type,"is currently not supported in this release.")
  }
}

qkid <- function(qkid, src="qkid", srcid=qkid, retrieved=yyyymmdd()) {
  structure(list(qkid=qkid, src=src, srcid=srcid, retrieved=retrieved), class="qkid")
}

is_qkid <- is.qkid <- function(x) inherits(x,"qkid")
to_qkid <- as.qkid <- function(x) {
  if(is_qkid(x))
    return(x)
  
  if(all(grepl(attr(.qkidVersion(x),"rx"),x))) {
    x <- qkid(as.character(x), "qkid", x)
    return(x)
  }
  stop("unable to convert to a valid qkid")
}

length.qkid <- function(x) {
  length(x$qkid)
}
classname <- function (qkid) {
  qkid <- to_qkid(qkid)
  unname(unlist(.QKIDCLASSMAP_V4[cls(qkid)]))
}

# quick extract of components
entity <- function(qkid) {
  qkid <- to_qkid(qkid)
  e <- substr(qkid$qkid, 0, 10)
  e
}
cls <- function(qkid) {
  qkid <- to_qkid(qkid)
  cls <- substr(qkid$qkid, 12, 15)
  cls
}
instrument <- function(qkid) {
  qkid <- to_qkid(qkid)
  i <- substr(qkid$qkid, 17, 25)
  i
}

detail <- function(qkid) {
  UseMethod("detail", qkid)
}
detail.default <- function(qkid) {
  stop("detail requires a qkid compatible object")
}
detail.qkid <- function(qkid) {
  structure(list(Entity=entity(qkid), Class=cls(qkid), Instrument=instrument(qkid)), class="qkid_detail")
}
# print methods
print.qkid_detail <- function(x, ...) {
  cat("QKID:\n")
  cat("  Entity:", x$Entity, "\n")
  cat("  Class:", x$Class, "\n")
  cat("  Instrument:", x$Instrument, "\n")
}


qk_ticker <- function(ticker, ...) {
  if(is.qkid(ticker))
    ticker <- to_ticker(ticker)
  .tickerMapping()
  id <- unclass(unname(unlist(mget(ticker, envir=.QKID_TICKER, ifnotfound=NA))))
  isna <- is.na(ticker)
  id[isna] <- NA
  qkid(id, "ticker", ticker)
}
to_ticker <- function(qkid, ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_ticker(qkid$qkid)$qkid) 
}

qk_cik <- function(cik, ...) {
  .cikMapping()
  if(is_qkid(cik))
    cik <- to_cik(cik)
  if(is.numeric(cik) || all(is.character(cik) & nchar(cik) <= 10)) 
    cik <- as.character(gsub("^0+","",cik))
  id <- unclass(unname(unlist(mget(cik, envir=.QKID_CIK, ifnotfound=NA))))
  isna <- is.na(id) | id == "N/A"
  if(any(isna)) {
    id[isna] <- NA
    id[is.na(id)] <- sprintf("%010s.0000.000000000",cik)
  }
  qkid(id, "cik", cik)
}
to_cik <- function(qkid, ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_cik(qkid$qkid)$qkid) 
}

qk_figi <- function(figi, type=c("figi","shareClass","composite"), ...) {
  .figiMapping()
  if(is_qkid(figi))
    figi <- to_figi(figi)
  id <- unclass(unname(unlist(mget(figi, envir=.QKID_FIGI, ifnotfound=NA))))
  isna <- is.na(figi)
  id[isna] <- NA
  qkid(id, "figi", figi)
}
to_figi <- function(qkid, type=c("figi","shareClass","composite"),  ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_figi(qkid$qkid)$qkid) 
}

qk_permid <- function(permid, type=c("org", "instrument", "quote"), ... ) {
  if(is_qkid(permid))
    permid <- to_permid(permid)
  permid <- as.character(permid)
  .permidMapping()
  id <- unclass(unname(unlist(mget(permid, envir=.QKID_PERMID, ifnotfound=NA))))
  isna <- is.na(permid)
  id[isna] <- NA
  qkid(id, "permid", permid)
}
to_permid <- function(qkid, type=c("org", "instrument", "quote"), ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_permid(qkid$qkid)$qkid) 
}

qk_fundname <- function(name, ...) {
  if(is_qkid(name))
    name <- to_name(name)
  name <- as.character(name)
  .fundnameMapping()
  id <- unclass(unname(unlist(mget(name, envir=.QKID_FUNDNAME, ifnotfound=NA))))
  qkid(id, "fundname", name)
}
to_fundname <- function(qkid, ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_name(qkid$qkid)$qkid) 
}

qk_name <- function(name, ...) {
  if(is_qkid(name))
    name <- to_name(name)
  name <- as.character(name)
  .nameMapping()
  id <- unclass(unname(unlist(mget(name, envir=.QKID_NAME, ifnotfound=NA))))
  isna <- is.na(name)
  id[isna] <- NA
  qkid(id, "name", name)
}
to_name <- function(qkid, ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_name(qkid$qkid)$qkid) 
}

# sector
qk_sector <- function(sector, ...) {
  if(is_qkid(sector))
    sector <- to_sector(sector)
  sector <- as.character(sector)
  if(length(sector) > 1)
    stop("qk_sector only accepts a single sector")
  .sectorMapping()
  id <- unclass(unname(unlist(mget(sector, envir=.QKID_SECTOR, ifnotfound=NA))))
  id <- na.omit(id)
  qkid(id, "sector", rep(sector, length=length(id)))
}
to_sector <- function(qkid, ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_sector(qkid$qkid)$qkid) 
}

# sic
qk_sic <- function(sic, ...) {
  if(is_qkid(sic))
    sic <- to_sic(sic)
  sic <- as.character(sic)
  if(length(sic) > 1)
    stop("qk_sic only accepts a single SIC code")
  .sicMapping()
  id <- unclass(unname(unlist(mget(sic, envir=.QKID_SIC, ifnotfound=NA))))
  id <- na.omit(id)
  qkid(id, "sic", rep(sic, length=length(id)))
}
to_sic <- function(qkid, ...) {
  stopifnot(is_qkid(qkid))
  as.character(qk_sic(qkid$qkid)$qkid) 
}



qk_isin <- function(isin, ...) {
  stop("isin conversion is not yet implemented.")
}
to_isin <- function(x, ...) {
  stop("isin conversion is not yet implemented.")
}
qk_cusip <- function(cusip, ...) {
  stop("cusip conversion is not yet implemented.")
}
to_cusip <- function(qkid, ...) {
  stop("cusip conversion is not yet implemented.")
}
qk_lei <- function(lei, ...) {
  stop("lei conversion is not yet implemented.")
}
to_lei <- function(qkid, ...) {
  stop("lei conversion is not yet implemented.")
}

c.qkid <- function(...) {
  ldots <- list(...)
  qkid(qkid=unlist(lapply(ldots, function(e) e$qkid)), 
       src=unlist(lapply(ldots, function(e) e$src)),
       srcid=unlist(lapply(ldots, function(e) e$srcid))
  )
} 
as.character.qkid <- function(x, ...) x$qkid

`[.qkid` <- function(x, i, ...) {
   qkid(x$qkid[i], src=x$src, srcid=x$srcid[i])
}
`[[.qkid` <- `[.qkid`

print.qkid <- function(x, ...) {
  if(all(x$src=="qkid")) {
    print(x$qkid,quote=FALSE)
  } else {
    print(setNames(x$qkid, nm=x$srcid),quote=FALSE)
  }
}

.onAttach <- function(libname, pkgname) {
  setHook(packageEvent("qkiosk", "attach"), function(...) {
    packageStartupMessage("\nWelcome to QUANTkiosk. View account details at https://quantkiosk.com/account")
    if( is.na(qk_get_apikey(req=FALSE)) )
      packageStartupMessage("\nQK API key is not set. See help(qk_set_apikey).")
  })
}

# fundamental data API access

print.qk_fn <- function(x, ...) {
  items <- unique(c(sapply(attributes(x)$qkid_item$data, names)))
  ids <- unique(names(attributes(x)$qkid_item$id))
  cat("\033[1;38;5;208mQK\033[22;39m Financials  ",length(ids)," x ",length(items),paste0("(",attributes(x)$view,")"),"\n")
  cat(.p("  \033[3mcompanies\033[0m: [", paste(head(ids),collapse=", "), ifelse(length(ids)>6,"...",""),"]\n"))
  cat(.p("  \033[3mitems\033[0m: [", paste(head(items),collapse=", "), ifelse(length(items)>6,", ...",""),"]\n"))
}

full.qk_fn <- function(x, showfull=TRUE, ...) {
  attr(x, "showfull") <- showfull
  x
}

qkiditems <- function(qkids,items,using=TRUE) {
  oqkids <- qkids
  oitems <- items
  if(length(qkids) < 1 || length(items) < 1)
    stop("requires at least one qkid and one item to be specified (e.g. qkids=qk_ticker('NVDA'), items='SALE')")
  if(!is.list(items))
    items <- list(items)
  if(length(qkids) > length(items)) 
    items <- rep(items, each=length(qkids))

  id_items <- sapply(setNames(1:length(qkids),nm=qkids$srcid), function(q) lapply(items[q], function(i) paste(entity(qkids[q]),i,sep='.')))
  data_items <- lapply(setNames(1:length(qkids), nm = sprintf("CIK%010d",as.integer(entity(qkids)))), function(q) sapply(items[[q]], function(i) {l <- list(NULL); l}))
  qkid_items <- list(id=id_items, data=data_items,qkids=oqkids,items=oitems,using=using)
  class(qkid_items) <- 'qkid_item'
  qkid_items
}
create_qkiditem <- function(fn, id, item) {
  qi <- attributes(fn)$qkid_items
  if(missing(id))
    id <- names(qi$id) 
  if(missing(item))
    item <- qi$items
  lapply(setNames(id,nm=id), function(id) {
    w_id <- which(!is.na(match(gsub("(.*?)[.].*[.](.*)", "\\1",unlist(lapply(names(qi$id), function(nm) paste(nm,qi$id[[nm]], sep = ".")))), id)))
    w_item <- which(!is.na(match(gsub("(.*?)[.].*[.](.*)", "\\2",unlist(lapply(names(qi$id), function(nm) paste(nm,qi$id[[nm]], sep = ".")))), item)))
    t(t(as.data.frame(qi$id)))[ intersect(w_id,w_item) ]
  })
}

print.qkid_item <- function(x, ...) {
  print(x$id)
}

str.qkid_item <- function(object, ...) {
  str(object$id)
}

.FNCODES <- new.env()
qk_fncodes <- function() {
  if(exists('codes',envir=.FNCODES))
    return(get('codes',envir=.FNCODES))
  fncodes <- .reqQKID("itemcodes/fn/codes")
  colnames(fncodes) <- c("stmt","code","desc")
  fncodes <- split(fncodes, fncodes$stmt)
  fncodes <- lapply(fncodes, function(.) { rownames(.) <- NULL; class(.) <- c('qk_df','data.frame'); . })
  assign('codes', fncodes, envir=.FNCODES)
  fncodes
}

qk_fn <- function(qkids,
               items,
               qkid_items=NULL,
               from=20000101,
               to=today(),
               asof=today(),
               asfiled=FALSE,
               aspit=getOption("qk.aspit",FALSE),
               ticker=TRUE,
               hide=TRUE, quiet=TRUE,
               cache=TRUE) {
  check_qkid(qkids)

  if(as.integer(from) > as.integer(to) || (as.integer(from) > 0 && as.integer(from) < 20000101))
    stop("'from' must be after 20000101 and before 'to' or a negative offset")

  view <- "asof"
  if(as.integer(asof) != as.integer(today()) || isTRUE(aspit))
    view <- "pit"
  if(isTRUE(asfiled))
    view <- "asfiled"

  if(is.null(qkid_items))
    qkid_items <- qkiditems(qkids, items, using=FALSE)

  fd <- .FD(qkid_items, view=view, ticker=ticker, hide=hide, quiet=quiet, cache=cache)
  elapsed <- attr(fd, "elapsed")
  oview <- view
  view <- attr(fd, "view")
  if(oview != view) {
    options(qk.aspit=FALSE)
    warning("Fundamental point-in-time view is not enabled for this account. Requests are converted to asof=today(). Run qk_account() for options.")
  }

  # reorder async calls to request order
  fd <- unlist(fd,recursive=FALSE)[names(unlist(qkid_items$data,recursive=FALSE))] 

  attributes(fd) <- list(elapsed=elapsed, qkid_items=qkid_items, view=view, showfull=FALSE)
  class(fd) <- c("qk_fn","list")
  return(fd)
}

.FD <- function(qkid_items, view=c('asof','asfiled','pit'), ticker=TRUE, hide=TRUE, quiet=TRUE, cache=FALSE) {
  view <- match.arg(view)
  stime <- Sys.time()

  data <- qkid_items$name

  arg_id_item <- qkid_items$id
  arg_id_item <- paste(unlist(arg_id_item), collapse=",")

  apiKey <- Sys.getenv("QK_API_KEY")
  if(apiKey == "")
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")

  if(isTRUE(cache)) {
    arg_id_item <- NULL
    for(qkid_item in unname(unlist(qkid_items$id))) {
      qkid_item <- strsplit(qkid_item,"\\.")[[1]]
      if( !is.null(.QKAPIDATA[["FD"]][[view]][[sprintf("CIK%010d",as.integer(qkid_item[1]))]][[qkid_item[2]]]) ) {
        data[[sprintf("CIK%010d",as.integer(qkid_item[1]))]][[qkid_item[2]]] <- .QKAPIDATA[["FD"]][[view]][[sprintf("CIK%010d",as.integer(qkid_item[1]))]][[qkid_item[2]]]
      } else {
        arg_id_item <- c(arg_id_item, paste(qkid_item, collapse="."))
      }
    }
    arg_id_item <- paste(arg_id_item, collapse=",")
  }

  if( nchar(arg_id_item) > 0) {
    req <- "https://api.qkiosk.io/data/fundamental"
    reqbody <- as.character(toJSON(list(apiKey=apiKey,id_item=arg_id_item,view=view,version=getOption("fn_version","")),auto_unbox=TRUE))
    if(!isTRUE(qkid_items$using)) {
      ids <- paste(as.character(as.integer(substring(qkid_items$qkids$qkid,1,10))),collapse=',')
      items <- paste(qkid_items$items,collapse=',')
      reqbody <- as.character(toJSON(list(apiKey=apiKey,ids=ids,items=items,view=view,version=getOption("fn_version","")),auto_unbox=TRUE))
    }
  
    reqstring <- req
    if(hide) reqstring <- gsub(apiKey,"XXXX", reqstring)
    if(!quiet) message("requesting: ",reqstring,"\n")
  
    # using POST
    handle = new_handle()
    handle_setheaders(handle, "Content-Type"="application/json")
    handle_setheaders(handle, `Accept-Encoding` = "gzip")
    handle_setopt(handle, customrequest = "POST")
    handle_setopt(handle, postfields = reqbody)
  
    resp <- curl_fetch_memory(req, handle)
    if(resp$status_code != "200")
      stop("permission denied - verify your API key is set",call.=FALSE)
    hdrs <- parse_headers(resp$headers, multiple=TRUE)
    res <- fromJSON(rawToChar(resp$content),simplifyVector=FALSE)
    if(res$View != view) {
      ## TODO: if asof was set and full pit is unavailable to subscriber, we need to warn
    }
    view <- res$View # downgraded from pit to asof if not available
  
    if(resp$status_code != 200)
      stop(paste0("error: unable to complete request: ",paste(res$Errors,collapse=",")), call.=FALSE)
  
    fd <- list()
    success <- function(res){
      fd <<- c(fd, list(res))
    }
    failure <- function(msg){
      warning(msg)
    }
  
    lapply(unlist(res$Urls), function(req) { h <- new_handle();handle_setopt(h,connecttimeout=30);curl_fetch_multi(req, success, failure,handle=h) } )
    multi_run()
  
    status_code <- lapply(fd, function(resp) { resp$status_code })
    content_type <- lapply(fd, function(resp) { resp$type })
    headers <- lapply(fd, function(resp) { parse_headers(resp$headers, multiple=TRUE) })
    # as.POSIXct(gsub("Last-Modified: ...,.","","Last-Modified: Tue, 04 Mar 2025 01:14:10 GMT"), format="%d %b %Y %H:%M:%S", tz="GMT")
    content <- lapply(fd, function(resp) { read.csv(textConnection(rawToChar(resp$content)), colClasses=c(fq='double',fytd='double',rptq='double',rpty='double')) })
    id_item <- lapply(fd, function(resp) gsub(".*/fundamental/[a-z]+?/(.*?)/(CIK.*)[.].*[?].*", "\\2,\\1",resp$url))
  
    idItem <- res$IdItem[match(sapply(fd, function(.) .$url), unlist(res$Urls))]
    idItems_id <- gsub("^0+","",sapply(idItem, `[[`, 1))
    idItems_item <- sapply(idItem, `[[`, 2)
  
    for(i in 1:length(content)) {
      if(status_code[[i]] != 200)
        content[[i]] <- data.frame()
      .QKAPIDATA[["FD"]][[view]][[sprintf("CIK%010d",as.integer(idItems_id[i]))]][[idItems_item[i]]] <- content[[i]]
      data[[sprintf("CIK%010d",as.integer(idItems_id[i]))]][[idItems_item[i]]] <- content[[i]]
    }
  }
  etime <- Sys.time() - stime

  attr(data, "elapsed") <- etime
  attr(data, "view") <- view
  data
}

addTTM <- function(x) {
  # add FP lagged fq and fytd - e.g. fq          -> fq
  #                                    [2022,Q1] ->   [2021,Q1]
  #x <- as.data.frame(x)
  is_interval <- TRUE
  if(all(is.na(x$fytd)))
    is_interval <- FALSE
  for(q in 1:4) x$fqpy[x$fqtr==q] <- c(NA,head(x[x$fqtr==q,'fq',drop=TRUE],-1))
  for(q in 1:4) x$fytdpy[x$fqtr==q] <- c(NA,head(x[x$fqtr==q,'fytd',drop=TRUE],-1))

  x$ann <- NA
  x$ann[x$fqtr==4] <- x[x$fqtr==4,ifelse(is_interval,'fytd','fq'),drop=TRUE]
  x$lastann <- locf(x[,'ann',drop=TRUE])

  x$ttm <- ifelse(is.na(x$ann), x$lastann + x$fq - x$fqpy, x$ann)
  if(!is_interval)
    x$ttm <- rollapply(x$fq, 4, sum, fill = NA, align = "right")/4
  x <- x[,-which(colnames(x)=='lastann')]
  x
}

pitCols <- c('cyqtr','cik','item','stmt','rstmt','iq','filed','fpe','fqtr','fq','fytd','ann','ttm','fqpy','fytdpy','fqd')
## client functions for PIT
pit_asof <- pitAsOf <- function(pit, dt=today(), qtrs=1) {
  if(inherits(pit, 'qk_fn')) {
    view <- attr(pit, 'view')
    return(as.qk_df(do.call(rbind, Filter(is.data.frame,lapply(pit, function(.) pit_asof(as.qk_df(., view=view), dt=dt, qtrs=qtrs)))),cls=c("qk_df_fn", "qk_df")))
  }
  if(qtrs < 0)
    qtrs <- length(unique(to_df(pit)$fpe))
  if(inherits(dt,"Date") || inherits(dt,"POSIXt")) 
    dt <- as.integer(format(dt,"%Y%m%d"))
  pit <- pit[as.integer(pit$filed) <= as.integer(dt),] # subset to only data available on `dt`
  if(nrow(pit) == 0)
    return(NULL)
  pit <- pit[rev(!duplicated(rev(pit$fpe))),] # find dups from latest
  if(attr(pit,"view") == "pit")
    pit <- addTTM(pit)
  rownames(pit) <- NULL
  #class(pit) <- c("qk_df_fn", "qk_df","data.frame")
  tail(pit, qtrs)
}
pit_asfiled <- pitAsFiled <- function(pit, dt=today(), qtrs=1) {
  if(inherits(pit, 'qk_fn')) {
    view <- attr(pit, 'view')
    return(as.qk_df(do.call(rbind, Filter(is.data.frame,lapply(pit, function(.) pit_asfiled(as.qk_df(., view=view), dt=dt, qtrs=qtrs)))),cls=c("qk_df_fn", "qk_df")))
  }
  if(qtrs < 0)
    qtrs <- length(unique(to_df(pit)$fpe))
  if(inherits(dt,"Date") || inherits(dt,"POSIXt")) 
    dt <- as.integer(format(dt,"%Y%m%d"))
  pit <- pit[as.integer(pit$filed) <= as.integer(dt),] # subset to only data available on `dt`
  if(nrow(pit) == 0)
    return(NULL)
  dups <- which(duplicated(pit$fpe))
  if(length(dups) > 0)
    pit <- pit[-dups,]
  if(nrow(pit) == 0)
    return(NULL)
  if(attr(pit,"view") == "pit")
    pit <- addTTM(pit)
  rownames(pit) <- NULL
  #class(pit) <- c("qk_df_fn", "qk_df","data.frame")
  tail(pit, qtrs)
}

# INTERNAL calls
.QKAPIDATA <- new.env(hash=TRUE)

.clearQKAPIDATA <- function() {
  rm(list=ls(.QKAPIDATA), envir=.QKAPIDATA)
}
.joinfd <- function(fd, freq=c('fq','fytd','ttm')) {
  freq <- match.arg(freq)
  fdts <- NULL
  for(id in fd) {
    for(item in id) {
      fdts <- cbind(fdts,xts(item[,freq],as.Date(as.character(item$cyqtr),format="%Y%m%d")))
    }
  }
  cnames <- paste(rep(names(fd),each=length(names(fd[[1]]))), names(fd[[1]]),sep='.')
  colnames(fdts) <- cnames
  fdts
}
locf <- function(x) {
  rs <- rle(x)
  v <- rs$values
  S <- NA
  for(i in seq_along(v)) {
    if(!is.na(v[i])) {
      S <- v[i] 
    } else 
    if(is.na(v[i]) && !is.na(S)) {
      v[i] <- S
    }
  }
  rs$values <- v
  x <- inverse.rle(rs)
  x
}

.qk_fn_display_cols <-
c("cik", "acceptance_time", "stmt", "item", "filed", "fpb", "fpe",
"fqd", "fp", "fqtr", "cyqtr", "fq", "fytd", "ttm", "ann", "rstmt"
)
print.qk_df_fn <- function(x, ...) {
  print.qk_df(x, ..., display_cols=.qk_fn_display_cols, title="QK Fundamentals Data Frame")
}

qk_audit <- function(x, ...) {
  UseMethod('qk_audit')
}
qk_audit.default <- function(x, row, accn, fpe, open=FALSE, ctx=-1, ...) { 
  if(missing(row) && nrow(x) == 1)
    row <- 1
  if(missing(row) && !is.null(attr(x, "highlight"))) {
    row <- attr(x,"highlight")
    if(is.numeric(row))
      row <- which(row >= 0)
    if(is.logical(row))
      row <- which(row)
  }
  if( missing(row) || length(row) > 1 || row > nrow(x) )
    stop("`row` must specify a single obs of 'x'")

  if(!missing(accn) && !missing(fpe))
    row <- which(x$accn == accn & x$fpe == fpe)

  audit <- list()
  audit$x <- x
  audit$ctx <- ctx
  audit$open <- open
  audit$row <- row
  audit$cik <- x$cik[row]
  audit$accn <- x$accn[row]
  audit$stmt <- x$stmt[row]
  audit$fpb <- x$fpb[row]
  audit$ffiled <- x$ffiled[row]
  audit$filed <- x$filed[row]
  audit$fpe2filed <- x$fpe2filed[row]
  audit$fpe <- x$fpe[row]
  audit$fqd <- x$fqd[row]
  audit$fytd <- x$fytd[row]
  audit$fq <- x$fq[row]
  audit$fytd <- x$fytd[row]
  audit$rpty <- x$rpty[row]
  audit$rptq <- x$rptq[row]
  audit$rptqd <- x$rptqd[row]
  audit$rptyd <- x$rptyd[row]
  audit$fqtr <- x$fqtr[row]
  audit$rstmt<- x$rstmt[row]
  audit$iq<- x$iq[row]
  audit$concept_id <- x$concept_id[row];

  audit$restated <- ""
  if(audit$ffiled < audit$filed) {
    audit$restated <- paste0("\033[48;5;1m\033[1;38;5;15m RESTATED \033[39m\033[49;0m \033[3;38;5;241m(initial filing on ",audit$ffiled,")\033[23;39;0m")
  }
#  cat("\n\033[1mFinancial Auditing\033[0m  ",restated,"\n\n")
#  cat("  filer:",to_name(qk_cik(cik)),"\n")
#  cat(paste0("  cik: ",cik,"  |  accn: ",accn,"  |  statement: ",stmt,"  |  filed: ",filed," (fiscal period to date filed: ",fpe2filed," days)\n"))
#  cat(paste0("  fiscal period (",fqtr,"):\n    fpb: ",fpb,"\n    fpe: ",fpe,"\n\n"))
#  cat(paste0("  fiscal qtr (",sprintf("% 3s",rptqd)," days):\t",sprintf("% 15s",format(fq  ,big.mark=",")),"\t(",ifelse(iq==1," imputed )",paste0(" reported: ",rptq," )")),"\n"))
#  cat(paste0("  fiscal ytd (",sprintf("% 3s",rptyd)," days):\t",sprintf("% 15s",format(fytd,big.mark=",")),"\t( reported: ",rpty," )\n\n"))
#  print.qk_df(x[row,rerow=FALSE], rerow=FALSE, maxwidth=min(150, getOption("width")))
#  cat("\n")
  audit$txt <- .audit_fn(audit$cik, audit$accn, audit$stmt, audit$concept_id, audit$ctx)

#  if(isTRUE(open)) {
#    browseURL(a$url)
#  } else {
#    cat(a$stmt,"\n")
#    if(restated != "") {
#      cat("\n\033[3;38;5;1m           * Restated values are filed in later quarters as either an amended form (e.g. 10-Q/A)\n             or in one of the previous period reference values displayed (or provided in XBRL) filing\033[0m.\n\n") 
#    }
#  }
  class(audit) <- "qkaudit"
  audit
}

print.qkaudit <- function(x, ...) {
  cat("\n\033[1mFinancial Auditing\033[0m  ",x$restated,"\n\n")
  cat("  filer:",to_name(qk_cik(x$cik)),"\n")
  cat(paste0("  cik: ",x$cik,"  |  accn: ",x$accn,"  |  statement: ",x$stmt,"  |  filed: ",x$filed," (fiscal period to date filed: ",x$fpe2filed," days)\n"))
  cat(paste0("  fiscal period (",x$fqtr,"):\n    fpb: ",x$fpb,"\n    fpe: ",x$fpe,"\n\n"))
  cat(paste0("  fiscal qtr (",sprintf("% 3s",x$rptqd)," days):\t",sprintf("% 15s",format(x$fq  ,big.mark=",")),"\t(",ifelse(x$iq==1," imputed )",paste0(" reported: ",x$rptq," )")),"\n"))
  cat(paste0("  fiscal ytd (",sprintf("% 3s",x$rptyd)," days):\t",sprintf("% 15s",format(x$fytd,big.mark=",")),"\t( reported: ",x$rpty," )\n\n"))
  print.qk_df(x$x[x$row,rerow=FALSE], rerow=FALSE, maxwidth=min(150, getOption("width")))
  cat("\n")
  if(isTRUE(x$open)) {
    browseURL(x$txt$url)
  } else {
    cat(x$txt$stmt,"\n")
    if(x$restated != "") {
      cat("\n\033[3;38;5;1m           * Restated values are filed in later quarters as either an amended form (e.g. 10-Q/A)\n             or in one of the previous period reference values displayed (or provided in XBRL) filing\033[0m.\n\n") 
    }
  }
  invisible(x) 
}

.audit_fn <- function (cik, accn, stmt, concept_id, ctx=-1) {
  apikey <- qk_get_apikey()
  req <- paste0('https://api.qkiosk.io/data/audit?id=',cik,'&req=',accn,'-',stmt,'%2bhash.txt&prd=audit/fn/stmt&apiKey=',apikey)
  resp <- curl_fetch_memory(req)
  if(resp$status_code != "200")
    stop("permission denied - verify your API key is set", call.=FALSE)
  L <- iconv(strsplit(rawToChar(resp$content),"\n")[[1]], "latin1", "ASCII", sub="")
  cids <- substr(L,6,37)
  header <- 1:(which(grepl("cid:",L))[1])
  L <- gsub("^.{38}", "", L)
  m <- match(concept_id,cids)
  L[-m] <- paste0("\033[90m", L[-m], "\033[0m")
  L[ m] <- paste0("\033[34;1m", L[m], "\033[39;0m")
  if( ctx > 0 ) {
    s <- max(1, (m + c(-ctx,ctx))[1])
    e <- min(length(L), (m + c(-ctx,ctx))[2])
    L <- L[ sort(unique(c(header, s:e))) ]
    #L <- L[ sort(unique(c(header,( (m-3):min(m+3,length(L)) )))) ]
    if((m-ctx) > length(header))
      L <- c(L[header],"              ---",L[-header])
  }
  L <- paste(L, collapse = "\n")
  url <- gsub(".*(http.*htm).*","\\1",L[1])
  list(stmt=L,url=url)
}

# new yyyymmdd class

today <- function() {
  yyyymmdd <- as.integer(format(Sys.Date(),"%Y%m%d"))
  class(yyyymmdd) <- "yyyymmdd"
  yyyymmdd
}

yyyymmdd <- function(x=Sys.time(), tz=Sys.getenv("TZ")) {
  if(any(is.numeric(x) & nchar(x) != 8)) stop("x must be of form YYYYMMDD")
  if(missing(tz))
    tz <- Sys.getenv("TZ")
  if(tz == "")
    tz <- "UTC"
  if(is.numeric(x)) {
    yyyymmdd <- as.integer(format(as.Date(as.POSIXlt(as.character(x), format="%Y%m%d", tz=tz), tz=tz),"%Y%m%d"))
  } else {
    yyyymmdd <- as.integer(format(as.Date(as.POSIXlt(x),tz=tz),"%Y%m%d"))
  }
  class(yyyymmdd) <- "yyyymmdd"
  yyyymmdd
}

is.yyyymmdd <- function(x) inherits(x, "yyyymmdd")

as.yyyymmdd <- function(x, ...) UseMethod("as.yyyymmdd", x)
as.yyyymmdd.yyyymmdd <- function(x, ...) x
as.yyyymmdd.numeric <- function(x, ...) {
  yyyymmdd(x)
}

as.yyyymmdd.Date <- function(x, ...) { x <- as.integer(format(x, "%Y%m%d")); class(x) <- "yyyymmdd"; x }

as.Date.yyyymmdd <- function(x, ...) {
  as.Date(as.character(x), format="%Y%m%d")
}

as.POSIXct.yyyymmdd <- function(x, ...) {
  as.POSIXct(as.character(x), format="%Y%m%d")
}

as.POSIXlt.yyyymmdd <- function(x, ...) {
  as.POSIXlt(as.character(x), format="%Y%m%d")
}


Ops.yyyymmdd <- function(e1,e2) {
  e2 <- as.integer(e2)
  e1 <- get(.Generic)(as.Date(e1),e2)
  as.yyyymmdd(e1)
}

print.yyyymmdd <- function(x, ...) {
 print(unclass(x))
}

seq.yyyymmdd <- function(from, to, by, length.out = NULL, along.with = NULL, ...) {
  if(!missing(to))
    to <- as.Date(to)
  if(!is.null(along.with))
    length.out <- length(along.with)
  if(missing(by))
    by <- 'day'
  s <- seq.Date(from=as.Date(from), to, by, length.out=length.out, ...)
  as.yyyymmdd(s)
}

diff.yyyymmdd <- function(x, lag = 1, differences = 1, ...) {
    diff(as.Date(x), lag=lag, differences=differences, ...)
}

qtrsback <- function(yyyyqq, N=1) {
  if(nchar(yyyyqq)==4)
    yyyyqq <- yyyyqq * 100
  if(N==1) 
    return(yyyyqq)
  N <- N - 1L
  z <- c(4,1,2,3)[(yyyyqq %% 100 - (1:N)) %% 4 + 1]
  zz <- z
  zz[z==4] <- -1
  zz[z<4] <- 0
  zz <- cumsum(zz)
  c(yyyyqq, (yyyyqq%/%100 + zz)*100 + z)
}

qtrs <- function(n) n * 91

days <- function(n) n

to_qq <- function(yyyymmdd) {
  UseMethod("to_qq")
}
to_yyyy <- function(yyyymmdd) {
  UseMethod("to_yyyy")
}
to_yyyyqq <- function(yyyymmdd) {
  UseMethod("to_yyyyqq")
}

to_qq.yyyymmdd <- function(yyyymmdd) ((as.integer(yyyymmdd) %% 10000 %/% 100)-1) %/% 3 + 1
to_yyyy.yyyymmdd <- function(yyyymmdd) as.integer(yyyymmdd) %/% 10000
to_yyyyqq.yyyymmdd <- function(yyyymmdd) as.integer(sprintf("%d0%d",to_yyyy(yyyymmdd), to_qq(yyyymmdd)))

from_to_yyyy_qq <- function(from,to) {
  s <- to_yyyyqq(seq.yyyymmdd(yyyymmdd(from),yyyymmdd(to)))
  yyyyqq <- s$yyyy*100 + s$qq # e.g. 202004
  u <- which(!duplicated(yyyyqq))
  list(yyyy=s$yyyy[u], qq=s$qq[u])
}

qk_clearcache <- function() {
  .clearQKAPIDATA()
  .clearQKID()
}
.clearQKID <- function() {
  rm(list=ls(.QKID_CIK), envir=.QKID_CIK)
  rm(list=ls(.QKID_TICKER), envir=.QKID_TICKER)
  rm(list=ls(.QKID_PERMID), envir=.QKID_PERMID)
  rm(list=ls(.QKID_FIGI), envir=.QKID_FIGI)
  rm(list=ls(.QKID_NAME), envir=.QKID_NAME)
}

.QKID_CIK <- new.env(hash=TRUE)
.cikMapping <- function() {
  if(length(.QKID_CIK)>0) return(TRUE)
  .ciks <- .reqQKID("cik")
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_CIK),.ciks$qkid,.ciks$cik)
  isprimary <- which(!substr(.ciks$qkid,15,15) %in% c("C","P"))
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_CIK),as.character(.ciks$cik[isprimary]),.ciks$qkid[isprimary])
  invisible(TRUE)
}

.QKID_TICKER <- new.env(hash=TRUE)
.tickerMapping <- function() {
  if(length(.QKID_TICKER)>0) return(TRUE)
  .tickers <- .reqQKID("ticker")
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_TICKER),.tickers$qkid,.tickers$ticker)
  isprimary <- which(!substr(.tickers$qkid,15,15) %in% c("C","P"))
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_TICKER),as.character(.tickers$ticker[isprimary]),.tickers$qkid[isprimary])
  invisible(TRUE)
}

.QKID_PERMID <- new.env(hash=TRUE)
.permidMapping <- function() {
  if(length(.QKID_PERMID)>0) return(TRUE)
  .permids <- .reqQKID("permid")
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_PERMID),.permids$qkid,.permids$permid)
  isprimary <- which(!substr(.permids$qkid,15,15) %in% c("C","P"))
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_PERMID),as.character(.permids$permid[isprimary]),.permids$qkid[isprimary])
  invisible(TRUE)
}
.QKID_FIGI <- new.env(hash=TRUE)
.figiMapping <- function() {
  if(length(.QKID_FIGI)>0) return(TRUE)
  .figis <- .reqQKID("figi")
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_FIGI),.figis$qkid,.figis$figi)
  isprimary <- which(!substr(.figis$qkid,15,15) %in% c("C","P"))
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_FIGI),as.character(.figis$figi[isprimary]),.figis$qkid[isprimary])
  invisible(TRUE)
}
.QKID_NAME <- new.env(hash=TRUE)
.nameMapping <- function() {
  if(length(.QKID_NAME)>0) return(TRUE)
  .names <- .reqQKID("name")
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_NAME),.names$qkid,.names$name)
  isprimary <- which(!substr(.names$qkid,15,15) %in% c("C","P"))
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_NAME),as.character(.names$name[isprimary]),.names$qkid[isprimary])
  invisible(TRUE)
}
.QKID_FUNDNAME <- new.env(hash=TRUE)
.fundnameMapping <- function() {
  if(length(.QKID_FUNDNAME)>0) return(TRUE)
  .names <- .reqQKID("fund")
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_FUNDNAME),.names$qkid,.names$fund)
  isprimary <- which(!substr(.names$qkid,15,15) %in% c("C","P"))
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_FUNDNAME),as.character(.names$fund[isprimary]),.names$qkid[isprimary])
  invisible(TRUE)
}

# One to Many
.QKID_SECTOR <- new.env(hash=TRUE)
.sectorMapping <- function() {
  if(length(.QKID_SECTOR)>0) return(TRUE)
  .sectors <- .reqQKID("sector")
  bysector <- split(.sectors, .sectors$sector)
  mapply(function(sector, companies) assign(sector, companies$qkid, envir = .QKID_SECTOR), names(bysector), bysector) 
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_SECTOR),.sectors$qkid,.sectors$sector)
  invisible(TRUE)
}
.QKID_SIC <- new.env(hash=TRUE)
.sicMapping <- function() {
  if(length(.QKID_SIC)>0) return(TRUE)
  .sics <- .reqQKID("sic")
  bysic <- split(.sics, .sics$sic)
  mapply(function(sic, companies) assign(sic, companies$qkid, envir = .QKID_SIC), names(bysic), bysic) 
  mapply(function(id1,id2) assign(id1,id2,envir=.QKID_SIC),.sics$qkid,.sics$sic)
  invisible(TRUE)
}

.reqQKID <- function(id,dt=NA) {
  apiKey <- Sys.getenv("QK_API_KEY")
  if(apiKey == "")
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")

  req <- "https://api.qkiosk.io/data/qkid"
  if(!is.na(dt)) {
    reqbody <- as.character(toJSON(list(apiKey=apiKey,ids=id,dt=dt),auto_unbox=TRUE))
  } else {
    reqbody <- as.character(toJSON(list(apiKey=apiKey,ids=id),auto_unbox=TRUE))
  }
  
  # using POST
  handle = new_handle()
  handle_setheaders(handle, "Content-Type"="application/json")
  handle_setopt(handle, customrequest = "POST")
  handle_setopt(handle, postfields = reqbody)
  
  resp <- curl_fetch_memory(req, handle)
  if(resp$status_code != "200") {
    warning(paste0("error retrieving ",id,":",resp$status_code), call.=FALSE)
    return(NULL)
  }
  hdrs <- parse_headers(resp$headers, multiple=TRUE)
  res <- fromJSON(rawToChar(resp$content),simplifyVector=FALSE)

  resp <- curl_fetch_memory(res$Urls[[1]])
  read.csv(textConnection(rawToChar(resp$content)), stringsAsFactors=FALSE)
}

subset.qk_fn <- function(x, ...) {
  x1 <- subset(x[], ...)
  as.qk_df(x1, view=attr(x,"view"), showfull=attr(x,"showfull"), cls=c("qk_df_fn","qk_df","data.frame"))
}

transform.qk_fn <- function(`_data`, ...) {
  x1 <- transform(`_data`[], ...)
  as.qk_df(x1, view=attr(`_data`,"view"), showfull=attr(`_data`,"showfull"))
}

qk_select <- function(x, ...) {
  UseMethod("qk_select")
}

qk_select.qk_fn <- function(x, id_item, id, item, i, j, ...) {
## List of 5
##  $ id   :List of 1
##   ..$ CROX: chr "0001334036.SALE"
##  $ data :List of 1
##   ..$ CIK0001334036:List of 1
##   .. ..$ SALE: NULL
##  $ qkids:Class 'qkid'  hidden list of 4
##   ..$ qkid     : chr "0001334036.0000.001SDGFR8"
##   ..$ src      : chr "ticker"
##   ..$ srcid    : chr "CROX"
##   ..$ retrieved: 'yyyymmdd' int 20250618
##  $ items: chr "SALE"
##  $ using: logi FALSE

  oattr <- attributes(x)
  qi <- attr(x, "qkid_items")$id
  if(!missing(id)) {
    ## this won't work
    items <- lapply(which(sapply(x, function(.) .$cik[1]) == id),function(.) x[[.]])
  } else
  if(!missing(item)) {
    items <- lapply(which(sapply(x, function(.) .$item[1]) == item),function(.) x[[.]])
  } else
  if(missing(id_item)) {
    items <- x
  } else {
    items <- lapply(id_item, function(id_item) {
      x[[match(id_item,gsub("(.*?)[.].*[.](.*)", "\\1.\\2", unlist(lapply(names(qi), function(nm) paste(nm, qi[[nm]], sep=".")))))]]
    })
  }
  attributes(items) <- oattr
  items
}
`[.qk_fn` <- to_df.qk_fn <- function(x, id_item, id, item, i, j, ...) {
  showfull <- attr(x, 'showfull')
  view <- attr(x, 'view')
  items <- qk_select(x, id_item, id, item, i, j, ...)
  x <- do.call(rbind, items)
  if(!missing(i) &&  missing(j)) {
    x <- x[i,,drop=FALSE]
  } else if( missing(i) && !missing(j)) {
    x <- x[,j,drop=FALSE]
  } else if(!missing(i) && !missing(j)) {
    x <- x[i,j,drop=FALSE]
  }

  rownames(x) <- NULL
  
  as.qk_df(x, showfull=showfull, view=view, cls=c("qk_df_fn","qk_df","data.frame"))
}

to_ts <- function(x, ...) {
  UseMethod("to_ts")
}
as.xts.qk_ts <- function(x, ...) {
  class(x) <- c('xts','zoo')
  x
}

to_ts.qk_fn <- function(x, i='fq', dt='fpe', ...) {
  to_ts.qk_df(to_df(x), i=i, dt=dt)
}

to_ts.qk_df <- function(x, i='fq', dt='fpe', ...) {
  x <- split(x, list(x$cik, x$item))
  xx <- lapply(x, function(x) {
    xi <- x[,i]
    dts <- x[,dt,drop=TRUE]
    xts(as.matrix(xi), as.Date(yyyymmdd(dts)))
  })
  cnames <- c(sapply(strsplit(names(xx),"\\."), function(.) paste(to_ticker(qk_cik(.[1])),.[2],i,sep='.')))
  x <- do.call(cbind,xx)
  colnames(x) <- cnames
  class(x) <- c("qk_ts","xts","zoo")
  x
}
print.qk_ts <- function(x, ...) {
  df <- as.qk_df(as.data.frame(x), showfull=TRUE)
  rownames(df) <- index(as.xts(x))
  print(df, rerow=FALSE)
}

to_df <- function(x, ...) {
  UseMethod("to_df")
}
to_df.qk_df <- function(x, ...) {
  x
}
to_df.data.frame <- function(x, ...) {
  class(x) <- c("qk_df",class(x))
  x
}


as.data.frame.qk_fn <- function(x, ...) {
  qk_select(x)
}

as.qk_df <- function(x, view="", showfull=FALSE, cls='qk_df', highlight=NULL) {
  if(!is.data.frame(x))
    stop('conversion to qk_df requires a data.frame. did you mean to_df?')
  class(x) <- unique(c(cls, "data.frame"))
  attr(x, "showfull") <- showfull
  attr(x, "view") <- view
  attr(x, "highlight") <- highlight
  x
}

full <- function(x, showfull=TRUE, ...) {
  UseMethod("full")
}
full.qk_df <- function(x, showfull=TRUE, ...) {
  attr(x, "showfull") <- showfull
  x 
}

head.qk_df <- function(x, ...) {
  cls <- class(x)
  sf <- attr(x, "showfull")
  v <- attr(x,"view")
  x <- head(as.data.frame(x), ...)
  hl <- attr(x,"highlight")
  if(!is.null(hl))
    hl <- head(hl, ...)
  rownames(x) <- NULL
  as.qk_df(x,view=v,showfull=sf,cls=cls, highlight=hl)
}
tail.qk_df <- function(x, ...) {
  cls <- class(x);
  sf <- attr(x, "showfull")
  v <- attr(x,"view")
  x <- tail(as.data.frame(x), ...)
  hl <- attr(x,"highlight")
  if(!is.null(hl))
    hl <- tail(hl, ...)
  rownames(x) <- NULL
  as.qk_df(x,view=v,showfull=sf,cls=cls, highlight=hl)
}
`[.qk_df` <- function(x, i, j, drop=FALSE, rerow=TRUE, ...) {
  cls <- class(x)
  x <- as.data.frame(x)
  sf <- attr(x, 'showfull')
  view <- attr(x, 'view')
  if(!missing(i) &&  missing(j)) {
    x <- x[i,,drop=drop]
  } else if( missing(i) && !missing(j)) {
    x <- x[,j,drop=drop]
  } else if(!missing(i) && !missing(j)) {
    x <- x[i,j,drop=drop]
  } else {
    sf <- attr(x, 'showfull')
  }
  if(rerow)
    rownames(x) <- NULL
  if(drop && NCOL(x)==1)
    return(x)
  as.qk_df(x, view=view,showfull=sf,cls)
}
subset.qk_df <- function(x, ...) {
  cls <- class(x)
  sf <- attributes(x)$showfull
  v <- attributes(x)$view
  hl <- attr(x,"highlight")
  rownames(x) <- 1:nrow(x)
  x <- as.data.frame(x)
  x <- subset(x, ...)
  attr(x,'showfull') <- sf
  attr(x,'view') <- v
  if(!is.null(hl)) {
    attr(x,'highlight') <- hl[as.integer(rownames(x))]
  }
  rownames(x) <- NULL
  class(x) <- cls
  x
}
transform.qk_df <- function(`_data`, ...) {
  cls <- class(`_data`)
  x1 <- transform.data.frame(`_data`[], ...)
  as.qk_df(x1, view=attr(`_data`,"view"), showfull=attr(`_data`,"showfull"),highlight=attr(`_data`,"highlight"),cls)
}

print.qk_df <- function(x, ..., maxwidth=getOption("width"), display_cols=NULL, digits=NULL, title="A QK Data Frame", topn=getOption("qkiosk.print.topn",5), nrows=getOption("qkiosk.print.nrows",100),rerow=TRUE, html=FALSE) {
  if(isTRUE(html)) {
    if(missing(maxwidth))
      maxwidth <- 170
    .to_html(x, maxwidth=maxwidth)
    return(invisible(x))
  }

  cat("\033[1;38;5;239m\033[3m",title,"\033[23m\033[22;39m\n")
  cat("\033[1;38;5;244m\033[3m","Data licensed by QUANTkiosk\u2122 without warranty. See terms for usage restrictions.","\033[23m\033[22;39m\n")

  if(rerow)
    rownames(x) <- NULL

  showfull <- attributes(x)$showfull
  highlight <- attributes(x)$highlight
  if( !all(display_cols %in% colnames(x)) )
    showfull <- TRUE

  if(nrow(x) > 0) {
    if( isTRUE(showfull) || is.null(display_cols) ) {
      cat(.style(x, highlight=highlight, topn=topn, nrows=nrows, maxwidth=maxwidth))
    } else {
      cat(.style(x[, display_cols], highlight=highlight, topn=topn, nrows=nrows, maxwidth=maxwidth))
      cat(" \033[3;38;5;244m* ",ncol(x)-length(display_cols),"hidden cols - use full() to view all",ncol(x),"columns\033[23;39m\n")
    }
  } else {
    print(data.frame())
  }
  invisible(x)
}

.fmtcol <- function(x, big.mark=getOption("qkiosk.big.mark",NA)) {
  if(is.character(x))
    x <- gsub("\t","  ",x)
  if(is.numeric(x) && is.na(big.mark))
    format(x,justify="right")
  else
    format(x)
} 
.p <- function(...) paste(..., sep="")

color_names <- c(yellow=228, green=155, blue=159, red=202)
highlight <- function(x, ...) {
  UseMethod("highlight")
}
highlight.data.frame <- function(x, i, color='yellow', bg, fg, ...) {
  if(is.character(color) && !color %in% names(color_names))
    stop("color must be one of '",paste(names(color_names)),"' or an ANSI rgb color (0-255)")
  if(is.character(color))
    color <- color_names[color]
  hl <- attr(x, "highlight")
  if(is.null(hl))
    hl <- rep(-1, len=nrow(x))
  mc <- match.call(expand.dots=FALSE)
  if(isTRUE(is.call(mc$i))) {
    i <- with(x,eval(mc$i))
    if(is.logical(i))
      i <- which(with(x, eval(mc$i), ...))
  }
  if(length(i) > 0)
    hl[i] <- color
  attr(x,"highlight") <- hl
  x
}

.theme <- list(rowid=list(fg="208"),fg=list(int="33",dbl="35",chr="235",na="248"),highlight=list(bg=228))

.to_html <- function(x, maxwidth=170) {
  if(requireNamespace("fansi", quietly = TRUE)) {
    ansi <- paste(capture.output(print(x,maxwidth=maxwidth)),collapse="\n")
    ansi <- gsub('"',"&quot;",gsub(" ", "&nbsp;", gsub("\n", "<br>", ansi)))
    #ansi <- html_esc(ansi)
    html <- fansi::to_html(ansi,warn=FALSE)
    html <- paste0("<div style='font-family:monospace;font-size:9pt;'>",html,"</div>")
    cat(html)
  } else {
    stop( "Package \"fansi\" must be installed to print using html.", call. = FALSE)
  }
}

.pipesep <- if(l10n_info()[["UTF-8"]]) { "\u2502" } else { "|" }
.style <- function(x, colsep=.pipesep, highlight, dig=4, theme=getOption("qkiosk.theme",.theme), maxwidth=getOption("width"), topn=getOption("qkiosk.df.topn",5), nrows=getOption("qkiosk.df.nrows",100),digits=NULL) {
  hl <- rep(-1, len=nrow(x))
  if( !missing(highlight) && !is.null(highlight) ) {
    hl <- highlight
  }
  if(length(dig) != ncol(x)) dig <- rep(dig[1],ncol(x))

  middle <- NA
  prows <- rows <- 1:nrow(x)
  if(nrows < nrow(x)) {
    rows <- c(1:topn,(nrow(x)-topn):nrow(x))
    prows <- c(1:topn,(nrow(x)-topn+1):nrow(x))
    middle <- topn+1
  }
  hl <- hl[rows]

  padding <- nchar(colsep)
  ncharna <- function(ch) ifelse(is.na(ch), 2, nchar(ch))
  max_w <- header <- styled <- NULL
  cnames <- sprintf("%-5s",colnames(x))
  rowchars <- max(3, nchar(rownames(x)))
  maxwidth <- maxwidth - 20
  header[[1]] <- paste0(rep(" ", rowchars), collapse="") # TODO: make QK
  for(j in 1:ncol(x)) {
     if(is.double(x[[j]]) && any(na.omit(x[[j]] %% 1 > 0))) # only apply to decimal precision
       x[[j]] <- signif(x[[j]],dig[j])
     max_w[j] <- max(ncharna(x[[j]][rows]), nchar(.fmtcol(cnames[j]))) + padding
     hdr_style <- sprintf("%%-%ds", max(3,ncharna(.fmtcol(x[[j]][rows])), nchar(cnames[j])))
     header[[j+1]] <- sprintf(hdr_style, cnames[j])
  }

  cbreaks <- cumsum(max_w) %/% as.integer(maxwidth)
  w_breaks <-split(1:(ncol(x)+1), c(0,cbreaks))
  w_breaks[[1]] <- w_breaks[[1]][-1]
  row_style <- sprintf("\033[38;5;%sm%% %ds\033[39m", theme$rowid$fg, max(3,nchar(rownames(x))))
  styled[[1]] <- sprintf(row_style, rownames(x)[rows])
  for(j in 1:ncol(x)) {
     fmt <- ifelse(is.character(x[[j]][rows]), .p("\033[3;38;5;",theme$fg$chr,"m%%-%ds\033[23;39m"), ifelse(is.integer(x[[j]][rows]), .p("\033[38;5;",theme$fg$int,"m%% %ds\033[39m"), .p("\033[38;5;",theme$fg$dbl,"m%% %ds\033[39m")))
     fmt <- ifelse(is.na(x[[j]][rows]), ifelse(is.character(x[[j]][rows]),"\033[3;38;2;248m%%-%ds\033[22;39m","\033[38;5;248m%% %ds\033[39m"), fmt)
     col_style <- sprintf(fmt, max(ncharna(x[[j]][rows]), nchar(cnames[j])))
     styled[[j+1]] <- sprintf(col_style, .fmtcol(x[[j]][rows]))
  }
  styled_rows <- rep(list(NULL), length(w_breaks))

  for(i in 1:length(w_breaks)) {
    styled_rows[[i]][1] <- paste0('\033[48;5;017m\033[38;5;231m',paste(sapply(header[c(1,w_breaks[[i]])], `[`, 1), collapse=colsep), '\033[39;0m\033[49;0m\n')
    ii <- 1
    for(row in 1:length(rows)) {

      if(row %% 2 == 0) {
        styled_rows[[i]][ii+1] <- paste0("\033[48;5;255m",paste(sapply(styled[c(1,w_breaks[[i]])], `[`, row), collapse=colsep), '\033[49;0m\n')
      } else {
        # terminals may view control characters as taking up space - pad exactly
        styled_rows[[i]][ii+1] <- paste0("\033[48;5;231m",paste(sapply(styled[c(1,w_breaks[[i]])], `[`, row), collapse=colsep), '\033[00;0m\n')
      }
      if(hl[row] >= 0 && hl[row] < 256)
        styled_rows[[i]][ii+1] <- paste0("\033[48;5;",hl[row],"m",paste(sapply(styled[c(1,w_breaks[[i]])], `[`, row), collapse=colsep), '\033[49;0m\n')

      if(!is.na(middle) && row==middle)
        styled_rows[[i]][ii+1] <- "  ----\n"

      ii <- ii + 1
    }
    # only show colnames at bottom if displaying non-truncated table that does not wrap
    #if(is.na(middle) && nrows < nrow(x) && length(w_breaks)==1 && getOption("qkiosk.df.colnames.after",TRUE))
    #  styled_rows[[i]][ii+1] <- paste0('\033[48;5;017m\033[38;1;097m',paste(sapply(header[c(1,w_breaks[[i]])], `[`, 1), collapse=colsep), '\033[39;0m\033[49;0m\n')
  }
  x <- paste0(unlist(styled_rows,recursive=FALSE), collapse='')
  attr(x,'styled') <- styled
  class(x) <- '.qkstyled'
  x
}

qk_complete <- function(dir,subscription='raw',rollup='5m',last='5m',date=today(),fmt=c('zip','tar.gz'),gzip=TRUE,hide=TRUE,quiet=FALSE,force=FALSE) {

  dir.create(dir, showWarnings=FALSE)
  # dir should be special directory created with qk.SetFilingsDir() -> class='qkiosk_filings_dir'
  fmt <- match.arg(fmt)

  apiKey <- Sys.getenv("QK_API_KEY")
  if(apiKey == "")
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")

  req <- sprintf("https://api.qkiosk.io/data/filings?fmt=%s&subscription=%s&rollup=%s&date=%s&last=%s&apiKey=%s",fmt,subscription,rollup,date,last,apiKey)
  reqstring <- req
  if(hide) reqstring <- gsub(apiKey,"XXXX", reqstring)
  if(!quiet) message("requesting: ",reqstring,"\n")

  resp <- curl_fetch_memory(req)
  res <- fromJSON(rawToChar(resp$content),simplifyVector=FALSE)
  files <- filings <- character(0)

  if(resp$status_code != 200)
    stop(resp$status_code," - ",res$Error)

  owd <- setwd(dir)
  on.exit(setwd(owd))

  ## TODO: add some sort of check maybe based on today and known rollup
  ## > seq(ISOdatetime(2024,6,28,17,9,0),ISOdatetime(2024,6,28,20,39,0),by=600)
  ## > as.POSIXlt(gsub("filings-(\\d{12})-.*$", "\\1", dir("edgar-testing/downloads/10m")),format="%Y%m%d%H%M")
  ##
  ## alternately could move logic into API endpoint where the Request sends list of rollups downloaded and returns missing ones

  dd <- file.path("downloads",rollup)
  dir.create(dd, showWarnings=FALSE, recursive=TRUE)
  urls <- unlist(res$Urls)
  filings_tofetch <- unlist(res$Filings)
  names(filings_tofetch) <- urls

  for(url in urls) {
    message("downloading ",filings_tofetch[url], "\n")
    if(!force && file.exists(file.path(dd, basename(filings_tofetch[url]))) ) {
      message(filings_tofetch[url], "already exists. use force=TRUE to redownload") 
      next
    }
    respurl <- curl_fetch_disk(url, tempfile())
    if(respurl$status_code != "200")
      stop("permission denied - verify your API key is set", .call=FALSE)
    hdrsurl <- parse_headers(respurl$headers, multiple=TRUE)
    filename <- gsub(".*filename=(.*?)$", "\\1", grep("filename", hdrsurl[[1]], value=TRUE))
    file.copy(respurl$content, file.path(dd,filename))
    message("copying ", filename, "\n")
    if(fmt=="zip") {
      files <- c(files, filename)
      newfilings <- unzip(file.path(dd,filename), list=TRUE)$Name
      filings <- c(filings, newfilings)
      message("extracting",length(newfilings),"filings ...")
      unzip(file.path(dd,filename), unzip="unzip", setTimes=TRUE)
      message("done.\n")
    }
  }
  if(gzip) {
    for(filing in filings) {
      filing <- c('-f', filing) # overwrite
      if(!quiet)
        filing <- c('-v',filing)
      system2(Sys.which('gzip'),args=filing)
    }
    filings <- paste(filings, "gz", sep=".")
  }
  message(length(filings),"filings extracted into",dir,"\n")
  invisible(list(downloads=files,filings=filings,dir=dir))
}

.QKID_UNIV <- new.env(hash=TRUE)

qk_univ <- function(univ,dt=NA,src="QK",cache=TRUE) {
  if( isTRUE(cache) && exists(univ, envir=.QKID_UNIV) ) {
    return(get(univ, envir=.QKID_UNIV))
  }
  if(src=="QK") {
    src <- "univ/QK"
    u <- .reqQKID(paste0(src,"/",univ,"/",univ),dt=dt)$qkid
  } else {
    stop("local universe files not supported yet")
  }
  if(is.null(u))
    stop("unable to retrieve universe")
  u <- qkid(u)
  assign(univ,u,envir=.QKID_UNIV)
  u
}

qk_account <- function(browser=FALSE, files=FALSE) {
  if(browser)
    return(browseURL("https://quantkiosk.com/account"))

  apiKey <- qk_get_apikey()

  nfiles <- 0L
  if(isTRUE(files))
    nfiles <- 1L

  req <- sprintf("https://api.qkiosk.io/account?apiKey=%s&nfiles=%d",apiKey,nfiles)
  resp <- curl_fetch_memory(req)
  if(resp$status_code != "200")
    stop("permission denied - verify your API key is set", call.=FALSE)

  account <- fromJSON(rawToChar(resp$content), simplifyVector=FALSE)
  if(isTRUE(files)) {
    names(account$DataFiles) <- sapply(account$DataFiles,function(.) .$File)
    names(account$UnivFiles) <- sapply(account$UnivFiles,function(.) .$File)
  }
  class(account) <- "qkaccount"
  account
}

qk_set_apikey <- function(apiKey, verbose=TRUE) {
  Sys.setenv(QK_API_KEY=apiKey)
  if(verbose)
    message("apiKey has been set in QK_API_KEY environment variable. Set in shell configuration or place this call in your .Rprofile")
  invisible(apiKey)
}

qk_get_apikey <- function(req=TRUE) {
  apiKey <- Sys.getenv("QK_API_KEY", unset=NA)
  if(req && (is.na(apiKey) || apiKey == ""))
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")
  apiKey
}

print.qkaccount <- function(x, ...) {
  cat("\nQUANTkiosk Account (",x$AsOf,"):\n\n")
  cat("  Daily Quota:",x$Quota," (Hard Quota:",x$HardQuota,")\n  Daily Usage",x$Usage,"\n")
  resets <- difftime(as.POSIXct(x$AsOf,format="%A, %d-%B-%y",tz='UTC') + 86400, as.POSIXct(x$AsOf,format="%A, %d-%B-%y %H:%M:%S",tz='UTC', units='hours'))
  cat("\n  Daily quota resets in: ", sprintf("%0.2f",resets)," hours\n")
  cat("\n  Your Files (use qk_download to view or download)\n\n")
  cat("    Data:",x$DataFilesN,"\n")
  cat("    Univs:",x$UnivFilesN,"\n")
  cat("\nVisit https://quantkiosk.com/account to change your plan or explore offerings.\n\n")
}

qk_download <- function(file=NULL, type = c("data", "univ", "trials"), n = 10, available=FALSE, retjson=TRUE) {
  x <- qk_account(files=TRUE)

  type <- c(data="DataFiles",univ="UnivFiles")[match.arg(type)]
  if(isTRUE(available)) {
    modified <- unname(sapply(x[[type]],function(.) .$LastMod))
    files <- unname(sapply(x[[type]],function(.) .$File))
    return(data.frame(modified,files))
  }
  if(!is.null(file)) {
    if(!file %in% names(c(x$DataFiles,x$UnivFiles)))
      stop(file,"is not available: use 'qk_download(available=TRUE) to view all files")
    url <- c(x$DataFiles,x$UnivFiles)[[file]]$Url
  } else {
    choice <- menu(sapply(x[[type]], function(.) sprintf("%s\t[%s]",
        .$File, .$LastMod))[1:n], title = sprintf("\nselect files to download (showing %d of %d)",n,x[[paste0(type,"N")]]))
    if (choice == 0)
        return()
    url <- x[[type]][[choice]]$Url
  }
  resp <- curl_fetch_memory(url)
  if(resp$status_code != "200")
    stop("download access not available", call.=FALSE)
  type <- gsub(".*[.](.*)$", "\\1", grep("Content-Disposition", parse_headers(resp$headers), value=TRUE))
  if(type == "csv") {
    dat <- as.qk_df(read.csv(textConnection(rawToChar(resp$content))))
  } else
  if(type == "psv") {
    dat <- as.qk_df(read.delim(textConnection(rawToChar(resp$content)),sep="|"))
  } else
  if(type == "json") {
    dat <- textConnection(rawToChar(resp$content))
    if(!isTRUE(retjson))
      dat <- as.qk_df(fromJSON(dat))
  }
  dat
}

qk_fund <- function() { stop("not yet implemented") }

qk_activist <- function(qkid, yyyyqq, qtrs=1, form=c("13D","13D13G"), wait=1, quiet=TRUE) {
  check_qkid(qkid)
  form <- match.arg(form)
  qk_beneficial(qkid,yyyyqq,qtrs,form=form,quiet=quiet,wait=wait)
}

qk_beneficial <- function(qkid, yyyyqq, qtrs=1, form=c("13D13G","13G","13D"), wait=1, quiet=TRUE) {
  check_qkid(qkid)
  apiKey <- Sys.getenv("QK_API_KEY")
  if(apiKey == "") 
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")

  by <- "filing"
  ep <- "ownership"

  retType <- "csv"

  form <- match.arg(form)
  #form <- switch(form, D="13D", G="13G", DG="13D13G", stop("invalid form type"))

  to_from <- qtrsback(yyyyqq,qtrs)
  yyyy <- sprintf("%04d",to_from %/% 100)
  yyyy <- unique(yyyy)
  qq <- rep(sprintf("%02d",0),length.out=length(yyyy))

  id <- entity(qkid)
  ticker <- to_ticker(qkid)

  vsn <- NA
  reqs <- 1
  totreqs <- length(qq)
  x <- lapply(1:totreqs, function(i) {
    message(sprintf("fetching %s (%s) for %s%s ...",ticker,id,yyyy[i],qq[i]))
    req <- sprintf("https://api.qkiosk.io/data/ownership?form=%s&apiKey=%s&ids=%s&aggType=a&yyyy=%s&qq=%s&by=%s&retType=%s",form,apiKey,id,yyyy[i],qq[i],by,retType)
    reqstring <- req
    reqstring <- gsub(apiKey,"XXXX", reqstring)
    if(!quiet) message("requesting: ",reqstring,"\n")

    resp <- curl_fetch_memory(req)
    if(reqs < totreqs) Sys.sleep(wait)
    if(resp$status != 200) {
      message(" unavailable.\n")  
      return(NULL)
    }
    vsn <<- gsub(paste0(".*ownership/(.*)/",form,".*"),"\\1",resp$url)
    h <- read.csv(textConnection(rawToChar(resp$content)))
    message("done.\n")
    h
  })

  x <- do.call(rbind, x)
  if(NROW(x) == 0) {
    x <- data.frame()
  }
  rownames(x) <- NULL
  attr(x,".vsn") <- vsn
  class(x) <- c("qk_df_beneficial","qk_df","data.frame")
  x
}
qk_insider <- function(qkid, yyyyqq, qtrs=1, form=c("insider","intent","sales"), wait=1, quiet=TRUE) {
  check_qkid(qkid)
  apiKey <- Sys.getenv("QK_API_KEY")
  if(apiKey == "") 
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")

  by <- "filing"
  ep <- "ownership"

  retType <- "csv"

  if(!is.numeric(qtrs)) stop("qtrs= must be integer number of qtrs to retrieve including current requested qtr")

  form <- match.arg(form)
  form <- c(insider="345",intent="144/intent",sales="144/sales")[form]
  to_from <- qtrsback(yyyyqq,qtrs)
  yyyy <- sprintf("%04d",to_from %/% 100)
  qq <- sprintf("%02d",to_from %% 100)

  id <- entity(qkid)
  ticker <- to_ticker(qkid)

  vsn <- NA
  reqs <- 1
  totreqs <- length(qq)
  x <- lapply(1:totreqs, function(i) {
    message(sprintf("fetching %s (%s) for %s%s ...",ticker,id,yyyy[i],qq[i]))
    req <- sprintf("https://api.qkiosk.io/data/ownership?form=%s&apiKey=%s&ids=%s&aggType=a&yyyy=%s&qq=%s&by=%s&retType=%s",form,apiKey,id,yyyy[i],qq[i],by,retType)
    reqstring <- req
    reqstring <- gsub(apiKey,"XXXX", reqstring)
    if(!quiet) message("requesting: ",reqstring,"\n")

    resp <- curl_fetch_memory(req)
    if(reqs < totreqs) Sys.sleep(wait)
    if(resp$status != 200) {
      message(" unavailable.\n")
      return(NULL)
    }
    vsn <<- gsub(paste0(".*ownership/(.*)/",form,".*"),"\\1",resp$url)
    h <- read.csv(textConnection(rawToChar(resp$content)))
    message("done.\n")
    h
  })

  x <- do.call(rbind, x)
  if(NROW(x) == 0) {
    x <- data.frame()
  }
  rownames(x) <- NULL
  attr(x,".vsn") <- vsn
  class(x) <- c("qk_df_insider", "qk_df","data.frame")
  x
}

qk_holders <- function(qkid, yyyyqq, qtrs=1, agg=TRUE, rollprev=TRUE, wait=1, quiet=TRUE) {
  check_qkid(qkid)

  if(to_yyyyqq(today()-qtrs(1)) == yyyyqq && isTRUE(rollprev)) {
    message(sprintf("rolling previous filers not yet reported into current data\n"),file=stderr())
    qtrs <- 2
    wait <- 0.1
  }

  apiKey <- Sys.getenv("QK_API_KEY")
  if(apiKey == "") 
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")

  if(nchar(yyyyqq) != 6 && !yyyyqq %% 100 %in% 1:4)
    stop("yyyyqq must be a exactly 6 digits and last two one of 01,02,03,04 representing quarter")

  to_from <- qtrsback(yyyyqq,qtrs)
  yyyy <- sprintf("%04d",to_from %/% 100)
  qq <- sprintf("%02d",to_from %% 100)

  use_entity <- FALSE
  id <- instrument(qkid)
  if(grepl("0[40]00",cls(qkid))) {
    id <- entity(qkid)
    use_entity <- TRUE
  }

  ticker <- to_ticker(qkid)

  vsn <- NA
  reqs <- 0
  totreqs <- length(qq)
  x <- lapply(1:totreqs, function(i) {
    message(sprintf("fetching %s (%s) for %s%s ...",ticker,id,yyyy[i],qq[i]),file=stderr())
    req <- sprintf("https://api.qkiosk.io/data/instrument?apiKey=%s&id=%s&yyyy=%s&qq=%s&entity=%s&agg=%s",apiKey,id,yyyy[i],qq[i],ifelse(use_entity,"true","false"),ifelse(agg,"true","false"))
    reqs <- reqs + 1
    rstart <- Sys.time()
    reqstring <- req
    reqstring <- gsub(apiKey,"XXXX", reqstring)
    if(!quiet) message("requesting: ",reqstring,"\n")

    resp <- curl_fetch_memory(req)
    if(reqs < totreqs && Sys.time() - rstart < wait) Sys.sleep(wait)
    if(resp$status != 200) {
      message("unavailable.\n")
      return(NULL)
    }
    vsn <<- gsub(".*ownership/(.*)/13f.*","\\1",resp$url)
    h <- read.csv(textConnection(rawToChar(resp$content)))
    h <- h[rev(order(h$value)),]
    message("done.\n")
    rownames(h) <- NULL
    h
  })

  re_sort <- FALSE
  if(rollprev && qtrs == 2 && length(x) == 2) {
    currentfilers <- unique(x[[1]]$filerCik)
    x[[2]] <- x[[2]][ !x[[2]]$filerCik %in% currentfilers, ]
    re_sort <- TRUE
  }

  x <- do.call(rbind, x)
  if(NROW(x) == 0) {
    x <- data.frame()
  }
  rownames(x) <- NULL
  attr(x, ".vsn") <- vsn
  class(x) <- c("qk_df_holders","qk_df","data.frame")

  if(re_sort) x <- qk_sort(x,by='value')

  x
}

qk_institutional <- function(qkid, yyyyqq, qtrs=1, agg=TRUE, wait=0, quiet=TRUE) {
  check_qkid(qkid)
  cik <- entity(qkid)

  apiKey <- Sys.getenv("QK_API_KEY")
  if(apiKey == "") 
    stop("apiKey not found in QK_API_KEY environment variable. Set in shell or use `Sys.setenv` from R.")

  by <- "filing"
  ep <- "ownership"

  retType <- "csv"

  if(nchar(yyyyqq) != 6 && !yyyyqq %% 100 %in% 1:4)
    stop("yyyyqq must be a exactly 6 digits and last two one of 01,02,03,04 representing quarter")

  to_from <- qtrsback(yyyyqq,qtrs)
  yyyy <- sprintf("%04d",to_from %/% 100)
  qq <- sprintf("%02d",to_from %% 100)
  aggType <- ifelse(agg, "a", "n")

  vsn <- NA
  reqs <- 1
  totreqs <- length(qq)
  x <- lapply(1:totreqs, function(i) {
    message(sprintf("fetching %s for %s%s ...",cik,yyyy[i],qq[i]))
    req <- sprintf("https://api.qkiosk.io/data/ownership?apiKey=%s&ids=%s&aggType=%s&yyyy=%s&qq=%s&by=%s&retType=%s",apiKey,cik,substr(aggType,1,1),yyyy[i],qq[i],by,retType)
    reqstring <- req
    reqstring <- gsub(apiKey,"XXXX", reqstring)
    if(!quiet) message("requesting: ",reqstring,"\n")

    resp <- curl_fetch_memory(req)
    if(reqs < totreqs) Sys.sleep(wait)
    if(resp$status != 200) {
      message("unavailable.\n")
      return(NULL)
    }

    vsn <<- gsub(".*ownership/(.*)/13f.*","\\1",resp$url)
    h <- read.csv(textConnection(rawToChar(resp$content)))
    message("done.\n")
    h
  })

  x <- do.call(rbind, x)
  if(NROW(x) == 0) {
    x <- data.frame()
  }
  rownames(x) <- NULL
  attr(x,".vsn") <- vsn
  class(x) <- c("qk_df_institutional","qk_df","data.frame")
  x
}

`[.qkinstitutional` <- function(x, i, j, ...) {
  x <- x[i, j, ...]
  class(x) <- c("qk_df_institutional","qk_df","data.frame")
  x 
}

qk_sort <- function(x, ...) {
  UseMethod("qk_sort", x)
}

#   $ shrsOrPrnAmtType      : chr  "SH" "SH" "SH" "SH" ...
#   $ portWgt               : num  0.00000767 0.00000016 0 0 0.00001008 ...
#   $ otherManager          : logi  TRUE TRUE TRUE TRUE TRUE TRUE ...
#   $ QtrsHeld              : int  10 9 2 8 32 25 2 43 38 6 ...
#   $ QOQSshPrnAmt          : int  -76849 0 -8330 -15628 -122106 -133246 0 -94100 -438296 0 ...
#   $ QOQValue              : int  -824292 -2400 -457734 -73139 -2688998 -1171121 -451524 -10440737 -3024452 -1961617 ...
#   $ QOQPortWgt            : 
`qk_sort.qk_df` <- function(x, by=c('value','shrsOrPrnAmtType','portWgt','QtrsHeld', 'qtrsHeld','QOQSshPrnAmt','QOQValue','QOQPortWgt'), incr=TRUE, ...) {
  by <- match.arg(by)
  if(!isTRUE(incr))
    rev <- function(x) x
  x[rev(order(x[,by,drop=TRUE])),]
}

qk_filter <- function(x, ...) {
  UseMethod("qk_filter", x)
}

qk_filter.qk_df <- function(x, by=c('top','incr','decr','new','del'), ...) {
  if(missing(by) || by == "top")
    return(x)

  by <- match.arg(by)

  if("incr" %in% by)
    x <- x[x[,'QOQValue'] > 0 & x[,'QOQSshPrnAmt'] > 0,]
  if("decr" %in% by)
    x <- x[x[,'QOQValue'] < 0 & x[,'QOQSshPrnAmt'] < 0,]
  if("new" %in% by)
    x <- x[x[,grep("newOrDel",colnames(x))] == "NEW",]
  if("del" %in% by)
    x <- x[x[,grep("newOrDel",colnames(x))] == "DEL",]

  x
}

qk_summary <- function(x, ...) {
  UseMethod("qk_summary", x)
}
summary_cols <- function(x) {
  if(inherits(x, "qk_df_institutional")) {
    cols <- c("issuer","titleOfClass","putCall","issuerSector","issuerTicker","shrsOrPrnAmt","QOQSshPrnAmt","value","QOQValue","portWgt","QOQPortWgt","QtrsHeld","hasOtherManager")
    cnames <- c("Name","Class","putCall","Sector","Symbol","SharesOrPrincipal","QOQ (SharesOrPrincipal)","Notional $","QOQ (Notional $)","Portfolio Wgt","QOQ (Port Wgt)", "Qtrs Held", "Other Mgrs")
  } else
  if(inherits(x, "qk_df_holders")) {
    cols <- c("filerName","filerCik","issuerName","issuerTitleOfClass","putCall","shrsOrPrnAmt","QOQSshPrnAmt","value","QOQValue","portWgt","QOQPortWgt","qtrsHeld","otherManager")
    cnames <- c("Filer","Filer CIK","Name","Class","putCall","SharesOrPrincipal","QOQ (SharesOrPrincipal)","Notional $","QOQ (Notional $)","Portfolio Wgt","QOQ (Port Wgt)", "Qtrs Held", "Other Mgrs")
  } else {
    stop("unknown data class:",class(x))
  }
  x <- x[,cols]
  colnames(x) <- cnames
  x
}
qk_summary.qk_df <- function(x, n=c(10,5), ...) {
  if(length(n) == 1)
    n <- c(n,n)
  top <- head(qk_sort(x,by='value'),n[1])
  incr <- head(qk_sort(qk_filter(x,by='incr'), by='QOQValue', incr=TRUE),n[2])
  decr <- head(qk_sort(qk_filter(x,by='decr'), by='QOQValue', incr=FALSE),n[2])
  new <- head(qk_sort(qk_filter(x,by='new'), by='QOQValue', incr=TRUE),n[2])
  del <- head(qk_sort(qk_filter(x,by='del'), by='QOQValue', incr=FALSE),n[2])

  list(Top10=summary_cols(top),Top5Increased=summary_cols(incr),Top5Decreased=summary_cols(decr),Top5New=summary_cols(new),Top5Deleted=summary_cols(del))
}

print.qk_df_institutional <- function(x, ...) {
  print.qk_df(x, ..., title="QK Institutional Ownership Data Frame", showfull=TRUE, display_cols=NULL)
}
print.qk_df_insider <- function(x, ...) {
  insider_cols <- c("filedDate", "acquiredOrDisposed", "txnDate", "issuerCik","issuerName","issuerTicker","ownerCik", "ownerName", "officerTitle", "isTenPct", "form", "isTxn", "txnCodeDesc", "txnAmount", "txnPrice", "txnValue", "secTitle", "ownershipForm", "postTxnShs")
  print.qk_df(x, ..., title="QK Insider Ownership Data Frame", showfull=TRUE, display_cols=insider_cols)
}
print.qk_df_holders <- function(x, ...) {
  print.qk_df(x, ..., title="QK Holders Data Frame", showfull=TRUE, display_cols=NULL)
}
print.qk_df_beneficial <- function(x, ...) {
  print.qk_df(x, ..., title="QK Ownership Data Frame", showfull=TRUE, display_cols=NULL)
}

\name{qkiosk-package}
\alias{qkiosk-package}
\alias{qkiosk}
\alias{qk_fund}
\alias{qk_clearcache}
\docType{package}
\title{
\packageTitle{qkiosk}
}
\description{
\packageDescription{qkiosk}
}
\details{

The DESCRIPTION file:
\packageDESCRIPTION{qkiosk}
\packageIndices{qkiosk}

Official QUANTkiosk R API
}
\author{
\packageAuthor{qkiosk}

Maintainer: \packageMaintainer{qkiosk}
}
\references{
\sQuote{http://quantkiosk.com}
}
\keyword{ package }

\name{qk_complete}
\alias{qk_complete}
\encoding{UTF-8}
\title{
Live Download EDGAR® Filings Raw & Processed Data
}
\description{
Provide access to QUANTkiosk™ \emph{QK Complete™} professional subscription. This function is intended
to be used as part of an ETL pipeline and run at fixed intervals to provide local live
capture of filings data.
}
\usage{
qk_complete(dir,
           subscription = "raw",
           rollup = "5m",
           last = "5m",
           date = today(),
           fmt = c("zip", "tar.gz"),
           gzip = TRUE,
           hide = TRUE,
           quiet = FALSE,
           force = FALSE)
}
\arguments{
  \item{dir}{
Directory to use for raw filings. Will be created if needed.
}
  \item{subscription}{
Subscription type varies with access. \sQuote{raw} retrieves
all \sQuote{.txt} filings submitted to the EDGAR® system.
}
  \item{rollup}{
Window of time for which new filings are provided.
}
  \item{last}{
Lookback window to query for \code{rollup} archives.
}
  \item{date}{
SEC date to query (America/New_York)
}
  \item{fmt}{
Return archive type. Default is \sQuote{zip} format which contains uncompressed (\sQuote{gzip}) raw
text filings.
}
  \item{gzip}{
Should files be compressed for local storage.
}
  \item{hide}{
Hide API key in request if \code{quiet=TRUE}.
}
  \item{quiet}{
Hide request from output.
}
  \item{force}{
Overwrite files if already downloaded.
}
}
\details{
\code{qk_complete} syncs live data subscriptions in near realtime. For most applications this is sufficient to maintain a full history
locally while the filings are arriving and being processed by the SEC. 
}
\value{
Called for its side-effects, the function returns a list of filings that have been extracted from the archive. The primary result is the
writing of the raw (or processed) filings to disk, as well as associated meta data created by the api.
}
\references{
SEC's EDGAR® Website: \sQuote{https://www.sec.gov/search-filings}
}
\author{
QUANTkiosk™
}
\note{
This endpoint requires both historical and live access. Consult your account page for details.
}

\examples{
\dontrun{
edgardir <- tempdir()
qk_complete(edgardir)

dir(edgardir)
}
}
\keyword{ data }
\concept{ filings }

\name{nke}
\alias{nke}
\encoding{UTF-8}
\docType{data}
\title{
Nike (NKE) Insider Ownership Data
}
\description{
Insider Ownership API example of large firm.
}
\usage{nke}
\format{
  A data frame with 60 observations on the following 40 variables.
  \describe{
    \item{\code{acceptanceDatetime}}{EDGAR® processing time}
    \item{\code{filing}}{canonical EDGAR® filing path}
    \item{\code{form}}{filing type: 3, 3/A (amended), 4, 4/A or  5, 5/A}
    \item{\code{is10b5One}}{indication if transaction was made pursuant to a contract, instruction or written plan}
    \item{\code{rptPeriod}}{a character vector}
    \item{\code{issuerCik}}{issuer Central Index Key}
    \item{\code{issuerName}}{issuer name}
    \item{\code{issuerTradingSymbol}}{issuer trading symbol}
    \item{\code{issuerQkid}}{issuer QKID™ symbol}
    \item{\code{ownerCik}}{insider Central Index Key}
    \item{\code{ownerName}}{a character vector}
    \item{\code{isDirector}}{is company director}
    \item{\code{isOfficer}}{is company officer}
    \item{\code{officerTitle}}{officer title}
    \item{\code{isTenPct}}{is a ten percent insider}
    \item{\code{secTitle}}{security title}
    \item{\code{secTitleFootnote}}{footnote for security title}
    \item{\code{conversionOrExercisePrice}}{price}
    \item{\code{conversionFootnote}}{a character vector}
    \item{\code{isTxn}}{is reporting line a transaction}
    \item{\code{isDeriv}}{is security a derivative}
    \item{\code{txnDate}}{transaction date}
    \item{\code{execDate}}{a logical vector}
    \item{\code{txnCode}}{transaction code pursuant to Rule 16b-3. See Details.}
    \item{\code{txnDesc}}{transaction footnote}
    \item{\code{txnAmount}}{transaction amount}
    \item{\code{acquiredOrDisposed}}{acquired or disposed}
    \item{\code{txnPrice}}{transaction price}
    \item{\code{txnValue}}{transaction value}
    \item{\code{exerciseDate}}{exercise date}
    \item{\code{exerciseDateFootnote}}{exercise date footnote}
    \item{\code{expirationDate}}{expiration date}
    \item{\code{expirationFootnote}}{expiration footnote}
    \item{\code{underlyingSecTitle}}{underlying security}
    \item{\code{underlyingShs}}{number of shares of underlying}
    \item{\code{postTxnShs}}{shares owned post transaction}
    \item{\code{postTxnFootnote}}{post transaction footnote}
    \item{\code{ownershipForm}}{form of ownership}
    \item{\code{ownershipNature}}{nature of ownership}
    \item{\code{ownershipFootnote}}{ownership footnote}
  }
}
\details{
The \sQuote{Insider Ownership} data provided by QUANTkiosk™ are gathered from SEC Form 3,4 and 5. The forms
are used for initial insider position, changes in position and exit of position, respectively.

These forms must be filed before the end of the second business day following
the day on which a transaction resulting in a change in beneficial ownership has been
executed (see references below).

Codes are as described below (direct from SEC)

\bold{General Transaction Codes}
\describe{
  \item{  P -}{Open market or private purchase of non-derivative or derivative security}
  \item{  S -}{Open market or private sale of non-derivative or derivative security}
  \item{  V -}{Transaction voluntarily reported earlier than required}
}

\bold{Rule 16b-3 Transaction Codes}
\describe{
  \item{  A —}{Grant, award or other acquisition pursuant to Rule 16b-3(d)}
  \item{  D —}{Disposition to the issuer of issuer equity securities pursuant to Rule 16b-3(e)}
  \item{  F —}{Payment of exercise price or tax liability by delivering or withholding securities incident to the receipt, exercise or vesting of a security issued in accordance with Rule 16b-3}
  \item{  I —}{Discretionary transaction in accordance with Rule 16b-3(f) resulting in acquisition or disposition of issuer securities}
  \item{  M —}{Exercise or conversion of derivative security exempted pursuant to Rule 16b-3}
}

\bold{Derivative Securities Codes (Except for transactions exempted pursuant to Rule 16b3)}
\describe{
  \item{  C —}{Conversion of derivative security}
  \item{  E —}{Expiration of short derivative position}
  \item{  H —}{Expiration (or cancellation) of long derivative position with value received}
  \item{  O —}{Exercise of out-of-the-money derivative security}
  \item{  X —}{Exercise of in-the-money or at-the-money derivative security}
}

\bold{Other Section 16(b) Exempt Transaction and Small Acquisition Codes (except for Rule 16b-3 codes above)}
\describe{
  \item{  C —}{Conversion of derivative security}
  \item{  G —}{Bona fide gift}
  \item{  L —}{Small acquisition under Rule 16a-6}
  \item{  W —}{Acquisition or disposition by will or the laws of descent and distribution}
  \item{  Z —}{Deposit into or withdrawal from voting trust}
}

\bold{Other Transaction Codes}
\describe{
  \item{  C —}{Conversion of derivative security}
  \item{  J —}{Other acquisition or disposition (describe transaction)}
  \item{  K —}{Transaction in equity swap or instrument with similar characteristics}
  \item{  U —}{Disposition pursuant to a tender of shares in a change of control transaction}
}
}
\references{
SEC Investor Bulletin: \sQuote{https://www.sec.gov/files/forms-3-4-5.pdf}

Form 4 Instructions: \sQuote{https://www.sec.gov/about/forms/form4data.pdf}
}
\examples{
data(nke)
}
\keyword{datasets}
\concept{Insider Ownership}
\concept{Ownership}

\name{account}
\alias{qk_account}
\alias{qk_download}
\alias{qk_set_apikey}
\alias{qk_get_apikey}
\encoding{UTF-8}
\title{
QUANTkiosk™ Account Access
}
\description{
Utilities for account usage and data downloads.
}
\usage{
qk_account(browser=FALSE, files=FALSE)

qk_download(file = NULL,
            type = c("data", "univ", "trials"),
            n = 10, available=FALSE, retjson = TRUE)

qk_get_apikey(req=TRUE)
qk_set_apikey(apiKey,verbose=TRUE)
}
\arguments{
    \item{browser}{
  open account page in browser
  }
    \item{files}{
  include custom files (if any) in account
  }
    \item{file}{
  file to download
  }
    \item{type}{
  data to download - either data or universe files.
  }
    \item{available}{
  return table of files and last modified times
  }
    \item{retjson}{
  should json be returned as raw string
  }
    \item{n}{
  number of files to select from
  }
    \item{req}{
  should an error be issued if not present
  }
    \item{apiKey}{
  your QUANTkiosk™ api key.
  }
    \item{verbose}{
  be verbose
  }
}
\details{
QUANTkiosk™ accounts control access as well as provide for custom data
access. 

To set (or get) your API key the functions \code{qk_get_apikey} and \code{qk_set_apikey}
are provided. These rely your environment variable, which may be set from with
your shell or from your session.

\code{qk_account} provides details on your current usage - as well as access to
any custom data files, universe files, or trial data you may have. To access the
files for download, the \code{qk_download} function allows for browsing your
remote data as well as downloading via an interactive menu or by specifying the name
of the file.
}
\value{
An invisible list containing:
  \item{AsOf}{Time of request} 
  \item{Quota}{Current daily quota in credits}
  \item{Usage}{Current usage in credits consumed}
  \item{DataFilesN}{Number of custom data files available}
  \item{DataFiles}{List of data files}
  \item{UnivFilesN}{Number of custom univ files available}
  \item{UnivFiles}{List of univ files}
}
\references{
QUANTkiosk™ Account Page: \sQuote{https://quantkiosk.com/account}
}
\author{
QUANTkiosk™ authors.
}
\keyword{ util }
\keyword{ web }
\concept{ QK Account Management }

\name{qk_univ}
\alias{qk_univ}
\encoding{UTF-8}
\title{
Universe Function
}
\description{
Tools to Fetch Universe Data
}
\usage{
qk_univ(univ, dt, src = "QK", cache=TRUE)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{univ}{
name of universe
}
  \item{dt}{
date for universe in form yyyymmdd. NOT IMPLEMENTED.
}
  \item{src}{
source of universe - \sQuote{QK} or local file path. NOT IMPLEMENTED.
}
  \item{cache}{
should data come from cache if available
}
}
\details{
The QK universe function allows for named universe requests. Each universe is constructed using a set of
rules.
}
\value{
A \code{qkid} object containing the universe.
}
\references{
\sQuote{https://quantkiosk.com/qkid}
}
\author{
QUANTkiosk™ team.
}
\note{
Neither \code{dt=} or \code{src=} are implemented in this version.
}

\seealso{
See Also \code{\link{qkid}}
}
\examples{
\dontrun{
qk100 <- qk_univ("QK100")

# top 1000 widely firms
qk1k <- qk_univ("QK1000")

# next 2000 most widely held firms
qk2k <- qk_univ("QK2000")
}

}
% Add one or more standard keywords, see file 'KEYWORDS' in the
% R documentation directory (show via RShowDoc("KEYWORDS")):
\keyword{ data }
\concept{ identifier }
\concept{ symbology }

\name{institutional}
\alias{institutional}
\alias{qk_institutional}
\alias{beneficial}
\alias{qk_beneficial}
\alias{activist}
\alias{qk_activist}
\alias{insider}
\alias{qk_insider}
\alias{holders}
\alias{qk_holders}
\encoding{UTF-8}
\title{
QUANTkiosk™ Ownership API Functions
}
\description{
Query all ownership across firms, invidividuals and
issuers.
}
\usage{
qk_institutional(qkid,
                 yyyyqq,
                 qtrs=1,
                 agg = TRUE,
                 wait = 0,
                 quiet = TRUE)

qk_insider(qkid,
           yyyyqq,
           qtrs=1,
           form = c('insider', 'intent', 'sales'),
           wait = 1,
           quiet = TRUE)

qk_beneficial(qkid,
              yyyyqq,
              qtrs=1,
              form = c('13D13G','13G','13D'),
              wait = 1,
              quiet = TRUE)

qk_activist(qkid,
            yyyyqq,
            qtrs=1,
            form = c('13D','13D13G'),
            wait = 1,
            quiet = TRUE)

qk_holders(qkid,
           yyyyqq,
           qtrs=1,
           agg = TRUE,
           rollprev = TRUE,
           wait = 1,
           quiet = TRUE)

}
\arguments{
  \item{qkid}{
a \code{qkid} object representing the \sQuote{QKID™} of reporting institution or filer. Use \code{qk_cik}, \code{qk_ticker}, or similar. See \code{qkid} for details.
}
  \item{yyyyqq}{
a number of the form \sQuote{YYYYQQ} where \sQuote{YYYY} is a 4-digit year, and \sQuote{QQ} is
a 2-digit quarter [00,01,02,03,04]. Use in conjuction with \code{qtrs} argument to request additional
prior quarters in one call. Four digit values will be treated as a full year (i.e. \sQuote{YYYY}) and expanded
to YYYY00.
}
  \item{qtrs}{
number of previous  quarters to request (including current \code{yyyyqq}.
}
  \item{form}{
for \code{qk_beneficial} and \code{qk_activist} requests, one of
\sQuote{13D} (activist), \sQuote{13G} (passive), \sQuote{13D13G} (activist + passive). For
\code{qk_insider} requests one of \sQuote{insider} (345), \sQuote{intent} or \sQuote{sales} (form 144).
}
  \item{agg}{
if \code{TRUE}, all other managers reporting are rolled up into one record, otherwise
filing will contain other managers if filed for as distinct holdings for one or more of the same
instruments.
}
  \item{rollprev}{
if \code{TRUE} and requested quarter is latest quarter to file, the previous completed quarter
will be used to augment the in-progress filings. Any new filers will override the last known
holdings positions, and yet-to-file updates will allow for last known positions to be seen.
Setting to \code{FALSE} will disable this merge and will only return the (potentially incomplete) quarter
being requested.
}
  \item{wait}{
time in seconds between calls.
}
  \item{quiet}{
make request silent.
}
}
\details{
QUANTkiosk™ Ownership combines multiple sources of information
in an custom formatted \code{data.frame}. At it's core, the data is used to understand
which participants have control over an entity. These participants file various forms
depending on who they are and what their intent may be.

Ownership reporting periods vary based on regulatory rules. Some are quarterly (e.g. Institutional and Funds),
whereas many are filed based on events (e.g. Insider, Beneficial/Activist). The arrival of the updates
is dependent on filer and SEC rules. All requests are specified for the quarter a filing was made.

To allow for consistent interface, \code{yyyyqq} is specified as the year \sQuote{yyyy} and the 2-digit quarter
being requested. For example, the second quarter of 2023 would be \code{yyyyqq=202302}. To request a full year,
one can use the special quarter \code{00}. To request all years use \code{yyyyqq=0} or \code{yyyyqq=000000}.

All usage is metered by data requested, e.g. four quarters or one year are equal in credits used.

For \sQuote{Institutional} (\code{qk_institutional}) aggregated (\code{agg=TRUE}) requests, we
roll up other managers filed within the main filer's report to better see total positions and
changes. If \code{agg=FALSE}, other managers are displayed as reported with quarter over quarter
change columns reflecting changes in exact reporting lines.
}

\section{Institutional}{
Institutional data consists of holdings reported by asset managers meeting certain criteria. Covering
more than 15,000 managers over the last decade, these are extracted in a point-in-time manner
from forms \sQuote{13F-HR}, \sQuote{13F-NT} and the amended versions of these forms.
}
\section{Insider}{
Insider ownership is reported through forms \sQuote{3}, \sQuote{4}, and \sQuote{5}. These are reports filed in
duplicate for insiders of a company to describe their ownership rights, including acquisition and disposal of
shares. These are filed after a change in ownership.
}
\section{Insider Intent}{
Insider Intent is a newer form, only required to be electronically filed in the last few years. This is used to indicate the intention to buy or sell, \emph{but not the
actual purchase or sale}. These are filed by \sQuote{insiders} of a company using form \sQuote{144}. This is split into \sQuote{intent} disclosures, as well as any
\sQuote{sale} in the 3 months prior to the filing.
}
\section{Holders}{
The holders data includes all owners of a particular company, allowing the cross sectional intitutional holdings in one call. This data is aggregated
from all individual reporting firms by quarter and made available as new filings arrive. At present this only contains \sQuote{13F-HR} reporting, updated
as filings arrive, including amendments to previous positions.
}
\section{Blockholders / Beneficial / Activist}{
Any owner with a position greater than 5\% must file under SEC's \sQuote{Schedule 13G (SC 13G)} or \sQuote{Schedule 13D (SC 13D)}. These are 
used to indicate control at or exceeding this threshold by one or more parties sharing control of the position.
}
\section{Funds}{
The largest of all assets under management comes from mutual and closed-end funds that report on form \sQuote{NPORT-P}. As opposed to
\sQuote{Institutional} filers, these represent publicly traded instruments. Reported quarterly.
}

\value{
A \sQuote{data.frame} of class \code{qkinstitutional} containing holdings or holders. See example data for fields.
}
\references{
SEC's EDGAR® Website: \sQuote{https://www.sec.gov/search-filings}

QK Ownership™: \sQuote{https://quantkiosk.com/ownership}

EDGAR® Form SC 13D, SC 13G FAQ: \sQuote{https://www.sec.gov/divisions/investment/13ffaq}

EDGAR® Form 13F-HR FAQ: \sQuote{https://www.sec.gov/divisions/investment/13ffaq}

EDGAR® Form 3,4,5 FAQ: \sQuote{https://www.sec.gov/divisions/investment/13ffaq}
}
\seealso{
Small filer example: \code{\link{sgcap}}

Large filer example with submanagers: \code{\link{deshaw}}

Holders example: \code{\link{crox}}

Benficial Owner example: \code{\link{pershing}}
}
\author{
QUANTkiosk™
}

\examples{
# all examples require an API key
\dontrun{
citadel <- qk_search_mgr("citadel")
citadel_h<- qk_institutional(citadel, yyyyqq=202301)

# all qtrs of 2023
citadel_h<- qk_institutional(citadel, yyyyqq=202300)

# nvidia's institutional holders for 202501, 202404, 202403, 202402
nvda_h <- qk_holders(qk_ticker("NVDA"), yyyyqq=202501, qtrs=4)

# nvidia's intent to sell 202501, 202404, 202403, 202402
nvda_insider<- qk_insider(qk_ticker("NVDA"), yyyyqq=202501, qtrs=4, form="insider")

# nvidia's intent to sell 202501, 202404, 202403, 202402
nvda_intent <- qk_insider(qk_ticker("NVDA"), yyyyqq=202501, qtrs=4, form="intent")

# nvidia's intent to sell 202501, 202404, 202403, 202402
nvda_sales <- qk_insider(qk_ticker("NVDA"), yyyyqq=202501, qtrs=4, form="sales")

# Restaurant Brands (QSR) block holders for 202501, 202404, 202403, 202402
qsr_b <- qk_beneficial(qk_ticker("QSR"), yyyyqq=2025)
}
}
\keyword{ data }
\concept{ QK Data }
\concept{ QK Institutional Ownership }

\name{yyyymmdd}
\alias{yyyymmdd}
\alias{is.yyyymmdd}
\alias{as.yyyymmdd}
\alias{qtrs}
\alias{days}
\alias{as.POSIXct.yyyymmdd}
\alias{as.POSIXlt.yyyymmdd}
\alias{as.Date.yyyymmdd}
\alias{as.yyyymmdd.Date}
\alias{as.yyyymmdd.numeric}
\alias{today}
\alias{to_qq}
\alias{to_yyyy}
\alias{to_yyyyqq}
\alias{seq.yyyymmdd}
\alias{diff.yyyymmdd}
\title{
Simple Dates
}
\description{
Simplified dates expressed as integers to facilitate
easy API access.
}
\usage{
yyyymmdd(x = Sys.time(), tz = Sys.getenv("TZ"))

today()

to_qq(yyyymmdd)
to_yyyy(yyyymmdd)
to_yyyyqq(yyyymmdd)

qtrs(n)
days(n)

\method{seq}{yyyymmdd}(from, to, by, length.out = NULL, along.with = NULL, ...)

\method{diff}{yyyymmdd}(x, lag = 1, differences = 1, ...)

is.yyyymmdd(x)

as.yyyymmdd(x, ...)

\method{as.yyyymmdd}{Date}(x, ...)
\method{as.yyyymmdd}{numeric}(x, ...)

\method{as.Date}{yyyymmdd}(x, ...)
\method{as.POSIXct}{yyyymmdd}(x, ...)
\method{as.POSIXlt}{yyyymmdd}(x, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
\R object
}
  \item{yyyymmdd}{
an 8-digit date
}
  \item{tz}{
time zone to use internally.
}
  \item{n}{
integer value to offset - may be positive or negative
}
  \item{from}{
starting date. Required
}
  \item{to}{
end date. Optional.
}
  \item{by}{
increment of the sequence. Optional. See \sQuote{Details}.
}
  \item{length.out}{
integer, optional. Desired length of the sequence.
}
  \item{along.with}{
take the length from the length of this argument.
}
  \item{lag}{
an integer indicating which lag to use.
}
  \item{differences}{
an integer indicating the order of the difference.
}
  \item{\dots}{
arguments passed to or from other methdod.
}
}
\details{
\R has a weath of date functionality but lacks a coherent integer
date representation that is capable of date math and proper timezone
handling. This set of tools provides this.
}
\value{
TBD
}
\references{
Most date functionality is provided by conversion to \R's \code{Date} internally and cast
back to a \code{yyyymmdd} before returning. See \code{\link{Date}} and \code{\link{seq.Date}} for more details.
}
\author{
QUANTkiosk
}
\seealso{
See also \code{\link{today}}
}
\examples{

# today - no TZ
today()
is.yyyymmdd(today())

# same output as today() but possible to change TZ
yyyymmdd()

# sequence
seq(yyyymmdd(), yyyymmdd() + 10)

seq(yyyymmdd(), by='month', length.out=10)
seq(yyyymmdd(), by='month', along.with=1:5)

# difference between dates
diff(yyyymmdd()+c(1:5,14:20), lag = 1, differences = 1)

# add durations
yyyymmdd() + 10
yyyymmdd() + qtrs(2)

# subtract durations
yyyymmdd() + qtrs(-2)
yyyymmdd() - qtrs(2)

# convert to yyyymmdd
as.yyyymmdd(Sys.Date())
as.yyyymmdd(20240124)

# convert back to R classes
as.Date(yyyymmdd())
as.POSIXct(yyyymmdd())

}
\keyword{ dates }
\concept{ dates }

\name{qk_select}
\alias{qk_select}
\alias{to_df}
\alias{to_ts}
\alias{full}
\alias{highlight}
\alias{highlight.data.frame}
\alias{as.qk_df}
\alias{to_ts.qk_fn}
\alias{to_ts.qk_df}
\alias{full.qk_fn}
\alias{full.qk_df}
\alias{subset.qk_fn}
\alias{subset.qk_df}
\alias{as.data.frame.qk_fn}
\alias{head.qk_df}
\alias{tail.qk_df}
\title{
Convert and Display QUANTkiosk Objects
}
\description{
Generic functions to extract, transform and display QUANTkiosk api data to common R objects.
}
\usage{
qk_select(x, ...)

to_df(x, ...)
as.qk_df(x, view="", showfull=FALSE, cls="qk_df", highlight=NULL)

to_ts(x, ...)
\method{to_ts}{qk_fn}(x, i='fq', dt='fpe', ...)
\method{to_ts}{qk_df}(x, i='fq', dt='fpe', ...)

full(x, showfull=TRUE, ...)
\method{full}{qk_fn}(x, showfull=TRUE, ...)
\method{full}{qk_df}(x, showfull=TRUE, ...)

\method{subset}{qk_fn}(x, ...)
\method{as.data.frame}{qk_fn}(x, ...)

\method{subset}{qk_df}(x, ...)
\method{head}{qk_df}(x, ...)
\method{tail}{qk_df}(x, ...)

highlight(x, ...)
\method{highlight}{data.frame}(x, i, color='yellow', bg, fg, ...)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
Object to transform. These are dispatched to object specific methods when appropriate.
}
  \item{view}{
internal view represented for use in pit functions
}
  \item{showfull}{
Display all hidden columns.
}
  \item{highlight}{
internal use
}
  \item{i}{
Row selection
}
  \item{dt}{
Column to use for indexing by time (to_ts)
}
  \item{color}{
background color for highlighting
}
  \item{bg}{
background color for highlighting. color names or ansi 256 colors.
}
  \item{fg}{
foreground color for highlighting. color names or ansi 256 colors.
}
  \item{cls}{
additional classes to add to conversion
}
  \item{\dots}{
Additional arguments depending on object class.
}
}
\details{
Most data fetched from QUANTkiosk results in either a \emph{QK} typed object or a \code{data.frame} of class \code{qk_df}.

Central to manipulation of these data, the package provides a variety of \code{s3} methods to interact with \pkg{base} or \pkg{stats} functionalty.
}
\value{A modified or transformed oject}

\examples{
head(crox)

highlight(crox, 2:3)

\dontrun{
# R >= 4.1.0
# crox |> highlight(2) |> highlight(3,'green')
}
}
\keyword{ manip }
\keyword{ data }
\concept{ formating }

\name{qk_fn}
\alias{qk_fn}
\alias{qk_fncodes}
\alias{qkiditems}
\alias{pit_asof}
\alias{pit_asfiled}
\alias{pitAsOf}
\alias{pitAsFiled}
\alias{qk_select.qk_fn}
\encoding{UTF-8}
\title{
Fundamentals Data
}
\description{
Functions to get \sQuote{as-reported} and \sQuote{revised} fundamentals line item data
from QK™/Financials API.
}
\usage{
qk_fn(qkids,
      items,
      qkid_items = NULL,
      from = 20000101,
      to = today(),
      asof = today(),
      asfiled = FALSE,
      aspit = getOption("qk.aspit",FALSE),
      ticker = TRUE,
      hide = TRUE,
      quiet = TRUE,
      cache = TRUE)

qk_fncodes()

\method{qk_select}{qk_fn}(x, id_item, id, item, i, j, ...)

qkiditems(qkids,items,using=TRUE)

pit_asof(pit,dt = today(), qtrs=1)

pit_asfiled(pit, dt = today(), qtrs=1)
}
\arguments{
  \item{qkids}{
a \code{qkid} vector representing the \sQuote{QKID™} of reporting institution or filer. Use \code{qk_cik}, \code{qk_ticker}, or similar. See \code{qkid} for details.
}
  \item{items}{
a vector or list of vectors containing QUANTkiosk™ item codes (e.g. SALE, NI, AT, ATC) corresponding to US-GAAP, DEI, or other line item(s). See qk_fncodes()
}
  \item{qkid_items}{
an object of class \code{qkid_item} created via \code{qkiditems}. if specified, will ignore \code{qkids} and \code{items}. If \code{qkid_items=NULL} (default) 
\code{qkids} and \code{items} will be converted into a corresponding \code{qkid_object}.
}
  \item{from}{
start date of returned series. 
}
  \item{to}{
end date of returned series. expresses as an integer or \code{yyyymmdd} date object.
}
  \item{asof}{
a number in form of YYYYMMDD. return last known values (incl. restated) for all periods up to \sQuote{asof} date. See details for more information on subscription limits.
}
  \item{asfiled}{
return as-filed values from original report (no restatements).
}
  \item{aspit}{
requests full point-in-time table for historical dates. ONLY AVAILABLE TO ENTERPRISE. By default this is attempted once and if not enabled
either the \emph{latest} point-in-time or as-filed data returned depening on \code{asof} or \code{asfiled} argument values.
}
  \item{ticker}{
should column names use tickers
}
  \item{hide}{
hide API Key from verbose (\code{quiet=FALSE}) output.
}
  \item{quiet}{
control verbosity.
}
  \item{cache}{
defaults to \code{TRUE}
}
  \item{pit}{
\code{qkid-item} timeseries from \code{qk_fn(..., aspit=TRUE)} call.
}
  \item{dt}{
date (yyyymmdd) used for point-in-time calculation. This will be converted using \code{as.yyyymmdd}.
}
  \item{qtrs}{
number of quarters to extract
}
  \item{x}{
an object of \code{qk_fn}
}
  \item{\ldots}{
additional arguments if applicable
}
  \item{id_item}{
a vector of id-items in form \sQuote{TICKER.ITEM} to extract \code{data.frame} from \code{qk_fn} object.
}
  \item{id}{
the ticker of the entity for extraction in \code{qk_select}. See \code{id_item} argument
}
  \item{item}{
a vector or list of vectors containing QUANTkiosk™ item codes (e.g. SALE, NI, AT, ATC) corresponding to US-GAAP, DEI, or other line item(s). See qk_fncodes()
}
  \item{using}{
internal flag to indicate input form.
}
  \item{i}{
extract rows from \code{qk_select} call.
}
  \item{j}{
extract colums from \code{qk_select} call.
}
}
\details{
To view current QUANTkiosk™ item codes call qk_fncodes(). This will return a list
organized by supported statements. This is constantly expanding.

For each item in \code{items} requested a request applied to every qkid in
\code{qkids}. All responses are cached by default. This can be disabled using \code{cache=FALSE}.

For enterprise subscriptions, if \code{view='pit'} is requested the subsequent object may contain restated and
as-filed values and times. Two functions
are provided to resolve to unique observations at each date: \code{pitAsFiled} converts the series to only include the initial
observation recorded, whereas \code{pitAsOf} reconciles restatements up to and including
the \code{date} specified. 
}
\value{
A multivariate \code{xts} timeseries if \code{full=FALSE}, containing one column for each \code{qkid}, \code{item} pair. The column names are derived from the QKID metadata. See examples for more details.

If \code{full=TRUE} a nested-list of data.frames are returned with additional details for each item value. See \code{pfe.sale} sample data. The general object looks like:

\itemize{
  \item qkid (e.g. "0001045810.000.001S5TZJ6" [NVDA])
    \itemize{
      \item sale (xts)
      \item eps (xts)
      \item atc (xts)
    }
  \item qkid (e.g. "0001564408.000.00441QMK5" [SNAP])
    \itemize{
      \item sale (xts)
      \item eps (xts)
      \item atc (xts)
    }
}

}
\references{
SEC EDGAR® filing and reporting rules.
}
\author{
QUANTkiosk™ team.
}
\note{
Depending on a user's subscription level, there may be restrictions on the universe of
available companies and available line items. For \code{asof} queries, the the requests
will resolve to the date of the query unless the subscription is for arbitrary point-in-time
data. See \sQuote{https://quantkiosk.com/point-in-time} for a full discussion of the details.
}

\seealso{
See Also \code{\link{qkid}}
See Also \code{\link{qk_fncodes}}
}
\examples{

\dontrun{
# get list of all QK™ codes
fncodes <- qk_fncodes()

# Pfizer and J&J quarterly revenue (SALE) and net income (NI) 
pfe <- qk_fn(qk_ticker(c("PFE","JNJ")), c("SALE","NI"))

tail( to_df(pfe) ) 
# single company, single item - Net Income
qkiditems(qk_ticker("AMD"), "NI")

# single company, multiple items - Revenue + Net Income
qkiditems(qk_ticker("AMD"), c("SALE","NI"))

# multiple companies, single item (used for each companies)
qkiditems(qk_ticker(c("AMD", "AAPL")), "NI")

# multiple companies, multiple items (used for each companies) - Assets + Assets Current
qkiditems(qk_ticker(c("AMD", "AAPL")), c("AT", "ATC"))

# multiple companies, multiple differnt items for each
qkiditems(qk_ticker(c("AMD", "AAPL")), list("NI",c("AT","ATC")))
}
}
\keyword{ data }
\concept{ fundamentals }
\concept{ financials }
\concept{ financial statements }

\name{sgcap}
\alias{sgcap}
\encoding{UTF-8}
\docType{data}
\title{
SG Capital Institutional Ownership Details (Aggregated)
}
\description{
Institutional Ownership API example of small manager filing.
}
\usage{sgcap}
\format{
  A data frame with 76 observations on the following 22 variables.
  \describe{
    \item{\code{filerName}}{name of filer (i.e. fund name)}
    \item{\code{filing}}{canonical EDGAR® filing path}
    \item{\code{submissionType}}{filing type: 13F-HR or 13F-HR/A (amended)}
    \item{\code{reportPeriod}}{period of report}
    \item{\code{filedDate}}{when the data was filed}
    \item{\code{issuer}}{name of instrument}
    \item{\code{titleOfClass}}{title or class name of instrument}
    \item{\code{issuerSIC}}{official standardized industrial code}
    \item{\code{issuerSector}}{two-character sector code from QUANTkiosk™}
    \item{\code{issuerTicker}}{common exchange ticker}
    \item{\code{issuerQkid}}{QKID™ symbol}
    \item{\code{value}}{fair market value of position at end of report period}
    \item{\code{shrsOrPrnAmt}}{number of shares held or principal amount}
    \item{\code{putCall}}{one of \sQuote{PUT}, \sQuote{CALL} or empty}
    \item{\code{shrsOrPrnAmtType}}{shares \sQuote{SH} or principal \sQuote{PRN}}
    \item{\code{portWgt}}{portfolio weight (QK™ calculated)}
    \item{\code{hasOtherManager}}{logical value indicating if submanagers are reported}
    \item{\code{QtrsHeld}}{quarters held (QK™ calculated)}
    \item{\code{QOQSshPrnAmt}}{quarter over quarter change in shares or principal amount of holding (QK™ calculated)}
    \item{\code{QOQValue}}{quarter over quarter change in value of holding (QK™ calculated)}
    \item{\code{QOQPortWgt}}{quarter over quarter change in portfolio weight (QK™ calculated)}
    \item{\code{newOrDel}}{is position \sQuote{NEW} or \sQuote{DEL}eted. See details. (QK™ calculated)}
  }
}

\details{
Institutional Ownership consists of quarterly \sQuote{13F-HR} and amended reports \sQuote{13F-HR/A} made by investors who hold at least $100 million of securities
declared by the U.S. Securities and Exchange Commission. Filers include
banks, insurance companies, and hedge funds. The filings are due 45 days after the end of each calendar quarter, and are filed
via the EDGAR® system. The securities reported must be from the official \sQuote{13(f) securities} list published by the SEC. See references section.

QUANTkiosk™ provides access to these filings in two forms at present. The first is an aggregated view
that includes one row per unique instrument held.
The second version (ex. \code{deshaw}) contains additional submanagers details. See the respective help page.

Additional fields not present in the official \sQuote{13F-HR} or \sQuote{13F-HR/A} but provided in both aggregated and non-aggregated views:
\itemize{
  \item issuerSIC
  \item issuerSector
  \item issuerQkid
  \item issuerTicker
}
These fields are useful for tracking changes between quarters as well as intrafiling analysis. 

Note that \code{hasOtherManager} is set to \code{TRUE} or \code{FALSE} if the filing reports other manager holdings. See \code{deshaw} for more details on additional fields in non-aggregated views

Additional fields pre-calculated by QUANTkiosk™ across time:
\itemize{
  \item QOQSshPrnAmt
  \item QOQValue
  \item QOQPortWgt
  \item newOrDel
  \item QtrsHeld
}
QK™ calculates these fields as quarter over quarter change in absolute terms. \code{newOrDel} marks holdings that are \sQuote{NEW} this quarter or \sQuote{DEL}eleted. Deleted entries are
not in the filing, but are provided to identify positions traded out of.  \code{QtrsHeld} shows duration of a continuously held position in aggregate, or in the case of submanager API data,
positions held continuously by the exact combination of submanagers. The latter is less reliable due to reporting considerations in the source data.
}

\references{
Ownership homepage: \sQuote{https://www.quantkiosk.com/ownership}

Ownership Dashboard: \sQuote{https://one.quantkiosk.com/ownership}

SG Capital Filings \sQuote{https://www.sec.gov/edgar/browse/?CIK=0001510099}

Form 13F-HR FAQ: \sQuote{https://www.sec.gov/divisions/investment/13ffaq}

Official 13(f) Securities \sQuote{https://www.sec.gov/divisions/investment/13flists}
}
\seealso{
R function: \code{\link{qk_institutional}}
Non-aggregated (Submanager) example: \code{\link{deshaw}}
}
\examples{
data(sgcap)
}
\keyword{datasets}
\concept{Institutional Ownership}

\name{qkid}
\alias{qkid}
\alias{.qkidVersion}
\alias{QKID}
\alias{qk_search}
\alias{qk_search_mgr}
\alias{qk_search_co}
\alias{qk_ticker}
\alias{to_ticker}
\alias{qk_cik}
\alias{to_cik}
\alias{qk_permid}
\alias{to_permid}
\alias{qk_figi}
\alias{to_figi}
\alias{qk_isin}
\alias{to_isin}
\alias{qk_cusip}
\alias{to_cusip}
\alias{qk_lei}
\alias{to_lei}
\alias{qk_sector}
\alias{to_sector}
\alias{qk_sic}
\alias{to_sic}
\alias{entity}
\alias{classname}
\alias{instrument}
\alias{cls}
\alias{detail}
\alias{qk_name}
\alias{to_name}
\alias{is_qkid}
\alias{to_qkid}

\alias{is.qkid}
\alias{as.qkid}
\alias{qk.E}
\alias{qk.Entity}
\alias{qk.C}
\alias{qk.Cls}
\alias{qk.I}
\alias{qk.Ticker}
\alias{as.Ticker}
\alias{qk.Figi}
\alias{as.Figi}
\alias{qk.Cik}
\alias{as.Cik}
\alias{qk.Cusip}
\alias{qk.Isin}
\alias{qk.Permid}
\alias{as.Permid}
\alias{qk.LEI}
\alias{qk.Name}
\alias{as.Name}
\encoding{UTF-8}
\title{
QKID™ Tools
}
\description{
Functions to create, extract and convert entity and instrument identifiers
within QUANTkiosk™'s QKID™ open symbology.
}
\usage{
qk_search(x, type=c('issuer','manager','fund','person'), n=10)
qk_search_co(x, n=10)
qk_search_mgr(x, n=10)

qk_ticker(ticker, ...)
to_ticker(qkid, ...)

qk_figi(figi, type = c("figi", "shareClass", "composite"), ...)
to_figi(qkid, type = c("figi", "shareClass", "composite"), ...)

qk_cik(cik, ...)
to_cik(qkid, ...)

qk_permid(permid, type = c("org", "instrument", "quote"), ...)
to_permid(qkid, type = c("org", "instrument", "quote"), ...)

qk_name(name, ...)
to_name(qkid, ...)

qk_sic(sic, ...)
to_sic(qkid, ...)

qk_sector(sector, ...)
to_sector(qkid, ...)

# convert string version of QKID to class `qkid`
qkid(qkid, src="qkid", srcid=qkid, retrieved=yyyymmdd())

is.qkid(x)
as.qkid(x)

.qkidVersion(qkid)
classname(qkid)
detail(qkid)

entity(qkid)
cls(qkid)
instrument(qkid)

}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{x}{
object to convert or test, or name to search
}
  \item{qkid}{
\code{qkid} classed object or one that can be converted to.
}
  \item{cik}{
a string representing a CIK (Central Index Key)

}
  \item{permid}{
a string representing a PermID

}
  \item{ticker}{
a string representing a ticker
}
  \item{figi}{
a string representing a FIGI
}
  \item{name}{
a string representing a name
}
  \item{sic}{
Standard Industrial Classification (SIC) code
}
  \item{sector}{
QK™ Sector (see below)
}
  \item{\dots}{
unused
}
  \item{type}{
optional conversion specifics, or one of \sQuote{issuer} or \sQuote{fund} in \code{qk_search_co} or \code{qk_search_mgr}, respectively.
}
  \item{src}{
source used in construction (e.g. \code{qk_ticker("AAPL")} => \sQuote{ticker})
}
  \item{srcid}{
value used in construction (e.g. \code{qk_ticker("AAPL")} => \sQuote{AAPL})
}
  \item{retrieved}{
date of retrieved (or created) qkid
}
  \item{n}{
number of search matches to return
}
}
\details{
\R functions to allow for mapping and manipulation of
QUANTkiosk™'s symbology as well as common identifiers using a special \code{qkid}
class.
}
\value{
QUANTkiosk™'s public identifier, known as \sQuote{QKID™} is the central
identifier used by the QK™ API. It is freely licensed under CC-BY-SA and
is designed to provide institutional quality symbology usable within
professional asset management while being simple to use.

For a full discussion, please refer to official website below and
the associated documentation. The \sQuote{QKID™} is a portmantaeu identifier,
combining well-established entity and instrument identifiers into a linked
symbology suitable for research and trading. There are three main components
to a fixed-width \sQuote{QKID™}:

\sQuote{QKID™} can also be mapped to sectors and SIC codes.

\itemize{
  \item ENTITY.CLS.INSTRUMENT
    \itemize{
      \item ENTITY: 10 characters uniquely matching an entity
      \item CLS: 4 character code indicating the instrument class
      \item INSTRUMENT: a unique, immutable instrument ID
    }
}

\sQuote{cls} and \sQuote{detail} provide access to \code{qkid} object
internals and details, returning broken-down identifiers or meta details about
the id.

To facilitate discovery, \pkg{qkiosk} provides search functionality. 
\code{qk_search_co} offers fuzzy matching to public companies (e.g. Pfizer or McDonalds),
returning a menu to select the match from. The return value is the matched \sQuote{QKID}.
Similar functionality is available to find investment managers (e.g. D.E Shaw) via
the \code{qk_search_mgr} function.

}
\references{
Official QKID™ site: \sQuote{https://www.quantkiosk.com/qkid}
}
\author{
QUANTkiosk™
}
\note{
Point-in-time QKID™ queries are currently not available outside of a subscription,
this may change in the future.
}

\examples{

# QKID™ string to class qkid
qkid("0000200406.0000.001S5SHQ9")

\dontrun{
qk_search_co("tesla")
qk_search_mgr("pershing")

jnj.qkid <- qk_ticker("JNJ")
to_ticker(jnj.qkid)
to_cik(jnj.qkid)

detail(jnj.qkid)


faang <- qk_ticker(c("META","AAPL","AMZN","NFLX","GOOGL"))
entity(faang)
cls(faang)
instrument(faang)

cls(faang)

to_ticker(faang)
to_name(faang)
to_permid(faang)
to_figi(faang)
to_cik(faang)
}

}
\keyword{ data }
\concept{ QKID }
\concept{ QKID Symbology }
\concept{ QKID Instrument }
\concept{ QKID Entity }

\name{qk_audit}
\alias{qk_audit}
\alias{qk_audit.default}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{
QUANTkiosk Data Auditing Tools
}
\description{
Methods to audit orginal data sources to verify data correctness and
interpretation.
}
\usage{
qk_audit(x, ...)

\method{qk_audit}{default}(x, row, accn, fpe, open = FALSE, ctx=-1, ...)
}
\arguments{
  \item{x}{
data.frame to audit
}
  \item{row}{
row of data.frame to display source
}
  \item{accn}{
optional accession number (\code{qk_fn})
}
  \item{fpe}{
optional fiscal period end (\code{qk_fn})
}
  \item{open}{
should browser window be opened to orginal source page
}
  \item{ctx}{
limit display of matching data to a smaller context window. defaults to full record.
}
  \item{\dots}{
unused
}
}
\details{
Non-market data is derived from myriad sources and contain billions of datum. To
better prove correctness (and to track data issues), QUANTkiosk provides a
user inteface to interrogate the original sources in a variety of ways.

The first application is for the fundamentals data we curate from billions
of datum published by companies. The \code{qk_audit} interface allows the
researcher to request the source data - \emph{in context} from the original
filings, if applicable.

For most of the standardized financial reporting line items (GAAP), we have
developed item codes to standardize reporting views when possible. These data
are coallesced from multiple places within filings including tabular data and
the \code{xbrl} (extensible business reporting language) mandaded in the US
and in other countries for many instances of reporting.

\code{qk_audit} is designed to be used regularly to inspect context as well
as numerical correctness of data returned by QUANTkiosk's APIs by using
extracted data from the original \code{HTML} filings. These original tables
do not include the additional contextual details present in the \code{qk_fn}
data - often only presenting a row label and a set of period to date values.
We process and extract the additional 30+ fields from a complex internal process
that includes validation steps as well as the very facility to inspect as presented
in \code{qk_audit}.
}
\value{
A displayed version of the text-formatted original html tables, including
context we are assigning and a highlighted section of the html that matches
the data in question.

Some values may we return in the API are actually imputed from the data
that may actually only report data in a period-to-date format.
}
\references{
\sQuote{www.sec.gov}
}
\author{
QUANTkiosk
}
\note{
This functionality is not all inclusive as locating data within a filings is not one for one. Often
data can be presented outside of statements or tables and therefore has no easy to display
data source. This may be possible in future version fo QK.
}

\seealso{
See Also as \code{\link{qk_fn}}
}
\examples{
\dontrun{
jnj <- qk_fn(qk_ticker("JNJ"), "SALE")

# pipes require R >= 4.1.0
# jnj |> to_df() |> qk_audit(55)
}
}
\keyword{ api }
\keyword{ data }
\concept{ fundamentals }
\concept{ financials }
\concept{ auditing }

\name{nvda_sales}
\alias{nvda_sales}
\encoding{UTF-8}
\docType{data}
\title{
Nvidia (NVDA) Restricted or Control Shares Sold
}
\description{
Corporate insider and affiliates intention to sell historical sales.
}
\usage{nvda_sales}
\format{
  A data frame with 386 observations on the following 18 variables.
  \describe{
    \item{\code{acceptanceDatetime}}{EDGAR® processing time}
    \item{\code{accn}}{internal EDGAR® accession number}
    \item{\code{filedDate}}{date filed in form yyyymmdd}
    \item{\code{filename}}{canonical EDGAR® filing path}
    \item{\code{form}}{filing type: 144 or 144/A}
    \item{\code{issuerCik}}{issuer Central Index Key}
    \item{\code{issuerName}}{issuer name}
    \item{\code{issuerQkid}}{issuer QKID™ symbol}
    \item{\code{filerCik}}{seller Central Index Key}
    \item{\code{filerName}}{seller name}
    \item{\code{relationshipToIssuer}}{a character vector}
    \item{\code{sellerName}}{name of seller}
    \item{\code{securitiesClassTitle}}{description of security}
    \item{\code{saleDateYYYYMMDD}}{date shares were sold}
    \item{\code{amountOfSecuritiesSold}}{number of shares sold}
    \item{\code{grossProceeds}}{total proceeds of sale in USD}
    \item{\code{asOf}}{unused}
    \item{\code{price}}{share price in USD}
  }
}
\details{
Form 144 is a requirement to disclose the intention to sell restricted shares. It is made up of two different
sections:

\describe{
  \item{Intent}{the proposed amount of shares expected to sell \code{nvda_intent} }
  \item{Sales}{all sales made over the preceeding three months \code{nvda_sales} }
}

The intent section of a form is unique across multiple forms filed by a person or entity. The \dQuote{Sales}
section may contain duplicate entries in the data provided by the API. Thsi is intentional as it represents
the time(s) the information was disclosed.

From the SEC, the securities and people reporting are:

Restricted securities are securities acquired in unregistered, private sales from the issuing company or from an affiliate of the issuer. Investors typically receive restricted securities through private placement offerings, Regulation D offerings, employee stock benefit plans, as compensation for professional services, or in exchange for providing "seed money" or start-up capital to the company. Rule 144(a)(3) identifies what sales produce restricted securities.

Control securities are those held by an affiliate of the issuing company. An affiliate is a person, such as an executive officer, a director or large shareholder, in a relationship of control with the issuer. Control means the power to direct the management and policies of the company in question, whether through the ownership of voting securities, by contract, or otherwise. If you buy securities from a controlling person or "affiliate," you take restricted securities, even if they were not restricted in the affiliate's hands.
}
\references{
Form 144 Investor Information: \sQuote{https://www.sec.gov/about/reports-publications/investorpubsrule144}
}
\examples{
data(nvda_sales)
}
\keyword{datasets}
\concept{Insider Ownership}
\concept{Ownership}

\name{pershing}
\alias{pershing}
\encoding{UTF-8}
\docType{data}
\title{
Pershing Square Beneficial and Activist Details
}
\description{
Beneficial and Activist API example of single manager block positions for 2019-2020
}
\usage{data("pershing")}
\format{
  A data frame with 72 observations on the following 25 variables.
  \describe{
    \item{\code{acceptanceDatetime}}{acceptance time to SEC's EDGAR® system}
    \item{\code{filing}}{canonical EDGAR® filing path}
    \item{\code{form}}{filing type: SC 13D, SC 13G, SC 13D/A or SC 13G/A }
    \item{\code{filedAsOf}}{when the data was filed}
    \item{\code{dateOfEvent}}{date reporting for}
    \item{\code{issuerName}}{name of issuer in submission}
    \item{\code{issuerCik}}{Central Index Key of issuer}
    \item{\code{issuerQkid}}{QKID™ symbol}
    \item{\code{issuerTicker}}{common exchange ticker}
    \item{\code{filerName}}{name of reporting filer}
    \item{\code{filerCik}}{Central Index Key of filer}
    \item{\code{rule13d13g}}{rule being reported for}
    \item{\code{namesOfRptPersons}}{individual or firm name who is reporting}
    \item{\code{isGrp}}{a character vector}
    \item{\code{srcOfFunds}}{a character vector}
    \item{\code{srcOfFundsDesc}}{a character vector}
    \item{\code{placeOfOrg}}{a character vector}
    \item{\code{soleVoting}}{a numeric vector}
    \item{\code{sharedVoting}}{a numeric vector}
    \item{\code{soleDispositoryPwr}}{a numeric vector}
    \item{\code{sharedDispositoryPwr}}{a numeric vector}
    \item{\code{aggAmtOwned}}{a numeric vector}
    \item{\code{pctOfClass}}{a numeric vector}
    \item{\code{personType}}{a character vector}
    \item{\code{personTypeDesc}}{a character vector}
  }
}
\details{
Beneficial and Activist Ownership describe owners (reporting) who maintain a large share of a public company (issuer). The forms
filed when a filer maintains or exceeds a 5\% are \sQuote{SC 13G} and \sQuote{SC 13D}, and the follow-on amendments as thos positions
change.

The difference in form types is related to filer \emph{intent}, specifically whether the 5\% owner is passive, or is
attempting to exert control over the company. The canonical example of an large investor who may file both is \sQuote{Pershing Square},
run by \sQuote{Bill Ackman}. We include this dataset to document the fields. Additional history is available for
all filings of this type as well as for all history.

Through 2024, all filings have come as unstructured text and as such are very difficult to systematically extract data. QUANTkiosk™
runs a complex automated extraction process to provide reliable data in a form that is made available in \R via
\code{qk_beneficial} and \code{qk_activist} functions. 
}
\references{
Ownership homepage: \sQuote{https://www.quantkiosk.com/ownership}

Pershing Square Filings: \sQuote{https://www.sec.gov/edgar/browse/?CIK=0001336528}

Form SC 13D/G information: \sQuote{https://www.sec.gov/rules-regulations/staff-guidance/compliance-disclosure-interpretations/exchange-act-sections-13d-13g-regulation-13d-g-beneficial-ownership-reporting}
}
\examples{
str(pershing)

\dontrun{
pershing.cik <- "0001336528"

qk_beneficial(pershing.cik, yyyy=c(2019,2020), qq=0)
}

}
\keyword{datasets}
\concept{Activist Ownership}
\concept{Beneficial Ownership}

\name{crox}
\alias{crox}
\encoding{UTF-8}
\docType{data}
\title{
Crocs Institutional Holders Details By Issuer
}
\description{
Institutional Holders API example that (tragically) never goes out of style.
}
\usage{crox}
\format{
  A data frame with 214 observations on the following 17 variables.
  \describe{
    \item{\code{issuerName}}{name of company}
    \item{\code{filing}}{canonical EDGAR® filing path}
    \item{\code{submissionType}}{filing type: 13F-HR or 13F-HR/A (amended)}
    \item{\code{reportPeriod}}{period of report}
    \item{\code{filedDate}}{when the data was filed}
    \item{\code{issuerTitleOfClass}}{reported title or class name}
    \item{\code{filerCik}}{Central Index Key of holder}
    \item{\code{filerName}}{name of holder}
    \item{\code{issuerQkid}}{QKID™ symbol}
    \item{\code{value}}{fair market value of position at end of period}
    \item{\code{QOQValue}}{quarter over quarter change in value of holding (QK™ calculated)}
    \item{\code{shrsOrPrnAmt}}{shares \sQuote{SH} or principal \sQuote{PRN}}
    \item{\code{QOQSshPrnAmt}}{quarter over quarter change in shares or principal amount of holding (QK™ calculated)}
    \item{\code{portWgt}}{portfolio weight (QK™ calculated)}
    \item{\code{QOQPortWgt}}{quarter over quarter change in portfolio weight (QK™ calculated)}
    \item{\code{qtrsHeld}}{quarters held (QK™ calculated)}
    \item{\code{newOrDeleted}}{is position \sQuote{NEW} or \sQuote{DEL}eted. See details. (QK™ calculated)}
  }
}
\details{
\code{crox} provides an aggregated view of holdings across all institutional filers. This is a cross-filer view
and is derived from between 5000-10000 individual filings.
}
\references{
Ownership homepage: \sQuote{https://www.quantkiosk.com/ownership}

Ownership Dashboard: \sQuote{https://api.quantkiosk.com/ownership}

Ownership API: \sQuote{https://www.quantkiosk.com/api/ownership}

SG Capital Filings \sQuote{https://www.sec.gov/edgar/browse/?CIK=0001510099}

Form 13F-HR FAQ: \sQuote{https://www.sec.gov/divisions/investment/13ffaq}

Official 13(f) Securities \sQuote{https://www.sec.gov/divisions/investment/13flists}
}
\seealso{
R function: \code{\link{qk_institutional}}
Aggregated Manager example: \code{\link{sgcap}}
Non-aggregated Manager (Submanager) example: \code{\link{deshaw}}
}
\examples{
data(crox)
}
\keyword{datasets}
\concept{Institutional Ownership}

\name{deshaw}
\alias{deshaw}
\encoding{UTF-8}
\docType{data}
\title{
D.E. Shaw Institutional Ownership Details (Including Submanagers)
}
\description{
Institutional Ownership API example of large manager filing with submanager details.
}
\usage{deshaw}
\format{
  A data frame with 6481 observations on the following 28 variables.
  \describe{
    \item{\code{filerName}}{name of filer (i.e. fund name)}
    \item{\code{filing}}{canonical EDGAR® filing path}
    \item{\code{submissionType}}{filing type: 13F-HR or 13F-HR/A (amended)}
    \item{\code{reportPeriod}}{period of report}
    \item{\code{filedDate}}{when the data was filed}
    \item{\code{inclMgrs}}{number of submanagers reported for. See \code{otherManager}}
    \item{\code{issuer}}{name of instrument}
    \item{\code{titleOfClass}}{title or class name of instrument}
    \item{\code{issuerSIC}}{official standardized industrial code}
    \item{\code{issuerSector}}{two-character sector code from QUANTkiosk™}
    \item{\code{issuerTicker}}{common exchange ticker}
    \item{\code{issuerQkid}}{QKID™ symbol}
    \item{\code{value}}{fair market value of position at end of report period}
    \item{\code{shrsOrPrnAmt}}{number of shares held or principal amount}
    \item{\code{putCall}}{one of \sQuote{PUT}, \sQuote{CALL} or empty}
    \item{\code{shrsOrPrnAmtType}}{shares \sQuote{SH} or principal \sQuote{PRN}}
    \item{\code{invDiscretion}}{type of discretion over investment decision}
    \item{\code{votingAuthSole}}{number of shares of \sQuote{SOLE} voting authority}
    \item{\code{votingAuthShared}}{number of shares of \sQuote{SHARED} voting authority}
    \item{\code{votingAuthNone}}{number of shares of \sQuote{NONE} voting authority}
    \item{\code{portWgt}}{portfolio weight (QK™ calculated)}
    \item{\code{hasOtherManager}}{logical indicating if record is for other manager}
    \item{\code{otherManager}}{reference number of other manager(s) from filing}
    \item{\code{otherManagerName}}{name of other manager(s) from filing}
    \item{\code{otherManagerFileNumber}}{file number of other manager(s)}
    \item{\code{QtrsHeld}}{quarters held (QK™ calculated)}
    \item{\code{QOQSshPrnAmt}}{quarter over quarter change in shares or principal amount of holding (QK™ calculated)}
    \item{\code{QOQValue}}{quarter over quarter change in value of holding (QK™ calculated)}
    \item{\code{QOQPortWgt}}{quarter over quarter change in portfolio weight (QK™ calculated)}
    \item{\code{newOrDel}}{is position \sQuote{NEW} or \sQuote{DEL}eted. See details. (QK™ calculated)}
  }
}
\details{
Institutional Ownership consists of quarterly \sQuote{13F-HR} and amended reports \sQuote{13F-HR/A} made by investors who hold at least $100 million of securities
declared by the U.S. Securities and Exchange Commission. Filers include
banks, insurance companies, and hedge funds. The filings are due 45 days after the end of each calendar quarter, and are filed
via the EDGAR® system. The securities reported must be from the official \sQuote{13(f) securities} list published by the SEC. See references section.

QUANTkiosk™ provides access to these filings in two forms at present. The first is an aggregated view
that includes one row per unique instrument held. This is described in full in the SG Capital example \code{\link{sgcap}}.
The second version (ex. \code{deshaw}) contains additional submanagers details:
\itemize{
  \item inclMgrs
  \item invDiscretion
  \item votingAuthSole
  \item votingAuthShared
  \item votingAuthNone
  \item otherManager
  \item otherManagerFileNumber
}

Additional fields not present in the official \sQuote{13F-HR} or \sQuote{13F-HR/A} but provided in both aggregated and non-aggregated views:
\itemize{
  \item issuerSIC
  \item issuerSector
  \item issuerQkid
  \item issuerTicker
}
These fields are useful for tracking changes between quarters as well as intrafiling analysis. \code{otherManager} and \code{otherManagerName} provides the reported reference number as filed
for each entry. Details of this requirement are beyond this help page, but reference the official SEC documentation and rules for clarity. QK™ does not merge
submanagers and instead reports the combined submanagers (if reported). For example, a 1;2 would mean the managers 1 and 2 share some aspect of the position.

\code{otherManagerFileNumber} is included to help disambiguate the reported ordering from the actual submanagers reporting. The names for the submanagers can
be had using the \code{meta=TRUE} option in \code{\link{qk_institutional}}.

Additional fields pre-calculated by QUANTkiosk™ across time:
\itemize{
  \item QOQSshPrnAmt
  \item QOQValue
  \item QOQPortWgt
  \item newOrDel
  \item QtrsHeld
}
QK™ calculates these fields as quarter over quarter change in absolute terms. \code{newOrDel} marks holdings that are \sQuote{NEW} this quarter or \sQuote{DEL}eleted. Deleted entries are
not in the filing, but are provided to identify positions traded out of.  \code{QtrsHeld} shows duration of a continuously held position in aggregate, or in the case of submanager API data,
positions held continuously by the exact combination of submanagers. The latter is less reliable due to reporting considerations in the source data.
}

\references{
Ownership homepage: \sQuote{https://www.quantkiosk.com/ownership}

Ownership Dashboard: \sQuote{https://api.quantkiosk.com/ownership}

Ownership API: \sQuote{https://www.quantkiosk.com/api/ownership}

D.E. Shaw Filings: \sQuote{https://www.sec.gov/edgar/browse/?CIK=0001009207}

Form 13F-HR FAQ: \sQuote{https://www.sec.gov/divisions/investment/13ffaq}

Official 13(f) Securities \sQuote{https://www.sec.gov/divisions/investment/13flists}
}
\seealso{
R function: \code{\link{qk_institutional}}
Aggregated example: \code{\link{sgcap}}
}
\examples{
data(deshaw)
}
\keyword{datasets}
\concept{Institutional Ownership}

\name{qk_filter}
\alias{qk_filter}
\alias{qk_sort}
\alias{qk_summary}
\alias{qk_filter.qk_df}
\alias{qk_sort.qk_df}
\alias{qk_summary.qk_df}
\title{
QK Object Manipulation
}
\description{
\sQuote{generic} functions to filter and sort data returned from the QK API. These are helper
functions designed to facilitate exploration of specific data structures provided by the API.
}
\usage{
qk_filter(x, ...)

\method{qk_filter}{qk_df}(x, by = c("top", "incr", "decr", "new", "del"), ...)

qk_sort(x, ...)
\method{qk_sort}{qk_df}(x, by = c("value", "shrsOrPrnAmtType", "portWgt",
                                  "QtrsHeld", "qtrsHeld", "QOQSshPrnAmt",
                                  "QOQValue", "QOQPortWgt"), incr = TRUE, ...)

qk_summary(x, ...)
\method{qk_summary}{qk_df}(x, n = c(10, 5), ...)
}
\arguments{
  \item{x}{
\R object
}
  \item{by}{
for \code{qk_sort} the column to sort on. For \code{qk_filter}, selection of rows matching description. See Details.
}
  \item{incr}{
sort in numeric increasing order. Set \code{FALSE} to order largest negative values. 
}
  \item{n}{
number of elements in tabular summary.
}
  \item{\dots}{
unused by current methods.
}
}
\details{
\pkg{qkiosk} makes available multiple data sets from the \emph{quantkiosk} platform. These are often information dense and require
further investigation. To help facilitate and align the experience across supported tools (e.g. dashboard, \R, python), a common
set of basic filtering is provided for convenience.

For \sQuote{Institutional Ownership} data returned from \code{qk_institutional}

See \code{\link{deshaw}} and \code{\link{sgcap}} for examples of data. 
}
\value{
filtered and sorted versions of original data.
}
\references{
See \sQuote{https://api.quantkiosk.com/ownership} dashboard examples
}
\author{
QUANTkiosk
}

\seealso{
See Also as \code{\link{qk_institutional}}, ~~~
}
\examples{

\donttest{
# SG Capital QKID using CIK 
sg <- qk_cik(1510099)

# sgcap is available as a dataset in package
sgcap <- qk_institutional(sg, yyyyqq=202401)
}

head(qk_filter(sgcap, by='incr'))
head(qk_sort(sgcap, by='QtrsHeld'))
head(qk_sort(qk_filter(sgcap, by='decr'), by='QOQValue', incr=FALSE))

qk_summary(sgcap)


}
\keyword{ data }
\concept{ summary }
\concept{ QK Ownership }

\name{pfe}
\alias{pfe}
\encoding{UTF-8}
\docType{data}
\title{
Pfizer (PFE) Revenue Data
}
\description{
QUANTkiosk™ Fundamentals API (As Of)
}
\usage{data("pfe")}
\format{
  A data frame with 60 observations on the following 40 variables.
  \describe{
    \item{\code{cik}}{the Central Index Key of the issuer}
    \item{\code{pik}}{the QKID™ Permanent Index Key of the issuer}
    \item{\code{acceptance_time}}{EDGAR® processing time}
    \item{\code{acceptance_utc}}{EDGAR® processing time in unix (seconds since 19700101)}
    \item{\code{stmt}}{statement including the item}
    \item{\code{item}}{the QK™ standardized item code}
    \item{\code{unit}}{unit of measurement - as reported}
    \item{\code{form}}{form filing type: 10-Q, 10-Q/A, 10-K, 10-K/A, 8-K, 8-K/A, 20-F, 20-F/A, 6-K, 6-K/A}
    \item{\code{filename}}{canonical EDGAR® filing path}
    \item{\code{accn}}{EDGAR® accession number}
    \item{\code{fye}}{fiscal year end}
    \item{\code{fyeq}}{fiscal yeat end quarter}
    \item{\code{fpb}}{fiscal period beginning}
    \item{\code{fpe}}{fiscal period end}
    \item{\code{filed}}{date filed}
    \item{\code{fpe2filed}}{days from fiscal period end to filed}
    \item{\code{filed40}}{forty day mark from fpe}
    \item{\code{ffiled}}{first filed - will be different in a restatment}
    \item{\code{cyqtr}}{calendar year quarter available}
    \item{\code{cyear}}{calendar year available}
    \item{\code{cqtr}}{calendar quarter available}
    \item{\code{rptq}}{value - fiscal quarter (reported)} 
    \item{\code{rptqd}}{reporting quarter days}
    \item{\code{rpty}}{value - fiscal year to date (reported)}
    \item{\code{rptyd}}{reporting year days}
    \item{\code{fp}}{fiscal period - e.g. Q1, Q2, Q3, Q4, FY}
    \item{\code{fqtr}}{fiscal quarter - e.g. 1, 2, 3, 4}
    \item{\code{fqd}}{fiscal quarter days}
    \item{\code{fq}}{value - calculated fiscal quarter}
    \item{\code{fytd}}{value - calculated fiscal year to date}
    \item{\code{ttm}}{value - calculated trailing twelve months}
    \item{\code{rstmt}}{restated flag}
    \item{\code{rstmt.q}}{restated flag - quarterly value}
    \item{\code{rstmt.y}}{restated flag - yearly value}
    \item{\code{iq}}{imputed quarterly}
    \item{\code{asof}}{as of date}
    \item{\code{concept_id}}{unique id for auditing}
    \item{\code{fqpy}}{fiscal quarter - prior year}
    \item{\code{fytdpy}}{fiscal year to day - prior year}
    \item{\code{ann}}{annual value - only fye}
  }
}
\details{
QUANTkiosk™ Fundamentals are extracted from original sources, using a systematic process. Data is extracted
and normalized into QK™ concepts that align across time and firms using a set of rules. Most concepts are
currently using \sQuote{US-GAAP} taxonomy. 
}
\source{
QUANTkiosk™ \sQuote{https://www.quantkiosk.com/fundamentals}
}
\examples{
to_df(pfe)
}
\keyword{datasets}
\concept{Fundamental Data}
\concept{Company Financials}
